#!/usr/bin/env python

import re
import logging

from ..exceptions import ParseError, GenerateError
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class MessagePacket(GenericPacket):
    """
    Class to represent APRS message packets.

    APRS message packets can be from one station to another, or general bulletins or announcements
    which are intended for a wider audience.

    See also APRS 1.01 C14 P71
    """

    def __init__(self, addressee: str = None, message: str = None, message_id: str = None,
                 bulletin_id: int = None, announcement_id: str = None,
                 group_bulletin_name: str = None, *args, **kwargs):
        """
        Create a new :class:`MessagePacket` object.

        :param str addressee: the message addressee
        :param str message: the message
        :param str message_id: an optional message ID
        :param int bulletin_id: the bulletin ID
        :param str announcement_id: the announcement ID
        :param str group_bulletin_name: the group bulletin name
        """
        super().__init__(*args, **kwargs)
        self._addressee = addressee
        self._message = message
        self._message_id = message_id
        self._bulletin_id = bulletin_id
        self._announcement_id = announcement_id
        self._group_bulletin_name = group_bulletin_name

        # Set the data type ID
        self.data_type_id = ":"

    @property
    def addressee(self) -> str:
        """Get the addressee of the message"""
        return self._addressee

    @addressee.setter
    def addressee(self, value: str):
        """Set the addressee of the message"""
        if value is None:
            # Clear the addressee
            self._addressee = None
        elif type(value) is str:
            if len(value) > 9:
                # The maximum length of the addressee is 9 characters
                raise ValueError(
                    "Addressee length cannot be longer than 9 characters ({} given)".format(
                        len(value)
                    )
                )
            else:
                self._addressee = value
        else:
            raise TypeError("Addressee must be of type 'str' ({} given)".format(type(value)))

    @property
    def message(self) -> str:
        """Get the message"""
        return self._message

    @message.setter
    def message(self, value: str):
        """Set the message"""
        if value is None:
            # Clear the message
            self._message = None
        elif type(value) is str:
            if len(value) > 67:
                # The maximum length of a message is 67 characters (C14 P71)
                logger.warning(
                    "Message length should not be longer than 67 characters ({} given)".format(
                        len(value)
                    )
                )
            self._message = value
        else:
            raise TypeError("Message must be of type 'str' ({} given)".format(type(value)))

    @property
    def message_id(self) -> str:
        """Get the message ID"""
        return self._message_id

    @message_id.setter
    def message_id(self, value: str):
        """Set the message ID"""
        if value is None:
            # Clear the message ID
            self._message_id = None
        elif type(value) is str:
            if len(value) > 5:
                # The maximum length of a message ID is 5 characters
                raise ValueError(
                    "Message ID length cannot be longer than 5 characters ({} given)".format(
                        len(value)
                    )
                )
            else:
                self._message_id = value
        else:
            raise TypeError("Message ID must be of type 'str' ({} given)".format(type(value)))

    @property
    def bulletin_id(self) -> int:
        """Get the bulletin ID"""
        return self._bulletin_id

    @bulletin_id.setter
    def bulletin_id(self, value: int):
        """Set the bulletin ID"""
        if value is None:
            # Clear the bulletin ID
            self._bulletin_id = None
        elif type(value) is int:
            if 0 <= value <= 9:
                self._bulletin_id = value
            else:
                # The bulletin must be 0 to 9
                raise ValueError(
                    "Bulletin ID must be a value between 0 and 9 inclusive ({} given)".format(
                        value
                    )
                )
        else:
            raise TypeError("Bulletin ID must be of type 'int' ({} given)".format(type(value)))

    @property
    def announcement_id(self) -> str:
        """Get the announcement ID"""
        return self._announcement_id

    @announcement_id.setter
    def announcement_id(self, value: str):
        """Set the announcement ID"""
        if value is None:
            # Clear the announcement ID
            self._announcement_id = None
        elif type(value) is str:
            if len(value) != 1:
                # An announcement ID is a single character
                raise ValueError(
                    "Announcement IDs must be a single character ({} given)".format(len(value)))
            else:
                self._announcement_id = value
        else:
            raise TypeError("Announcement ID must be of type 'str' ({} given)".format(type(value)))

    @property
    def group_bulletin_name(self) -> str:
        """Get the group bulletin name"""
        return self._group_bulletin_name

    @group_bulletin_name.setter
    def group_bulletin_name(self, value: str):
        """Set the group bulletin name"""
        if value is None:
            # Clear the group bulletin name
            self._group_bulletin_name = None
        elif type(value) is str:
            if len(value) > 5:
                # The maximum length of a group bulletin name is 5 characters
                raise ValueError(
                    "Group bulletin names cannot be longer than 5 characters ({} given)".format(
                        len(value)
                    )
                )
            else:
                self._group_bulletin_name = value
        else:
            raise TypeError("Group bulletin name must be of type 'str' ({} given)".format(
                type(value))
            )

    def _parse(self) -> bool:
        """
        Parse a message packet.

        The parsed values are stored within the current object.
        """

        # If this is a message, then ':" MUST be in the 9th position (C14 P71)
        try:
            if self._info[9] != ":":
                raise ParseError("Invalid message packet (missing : in 9th position)", self)

        except IndexError:
            raise ParseError("Invalid message packet (packet is too short)")

        # Split the message into the addressee and the actual message
        addressee = self._info[0:9]
        self.addressee = addressee.rstrip()
        message = self._info[10:]

        logger.debug("Message is addressed to {}, message is {}".format(addressee, message))

        # Is this a bulletin/announcement?
        if addressee[0:3] == "BLN":
            logger.debug("Message is a bulletin")

            if re.match("[0-9]", addressee[3]):
                # Bulletins have the format BLNn or BLNnaaaaa, where n is a digit between 0 and 9
                # and aaaaa is an optional group bulletin identifier
                if addressee[4:9] == "     ":
                    # Bulletin
                    self.bulletin_id = int(addressee[3])
                    logger.debug("Bulletin {}".format(self.bulletin_id))
                else:
                    # Group bulletin
                    self.group_bulletin_name = addressee[4:9].rstrip()
                    self.bulletin_id = int(addressee[3])

                    logger.debug("Group bulletin {} ({})".format(
                        self.group_bulletin_name, self.bulletin_id
                    ))

            elif re.match("[A-Z]", addressee[3]):
                # Announcements have the format BLNa, where a is a character between A and Z
                if addressee[4:9] == "     ":
                    # Announcement
                    self.announcement_id = addressee[3]
                    logger.debug("Announcement {}".format(self.announcement_id))
                else:
                    # Incorrectly-formatted bulletin
                    raise ParseError(
                        "Incorrectly-formatted announcement: {}".format(addressee), self
                    )

            else:
                # Incorrectly-formatted bulletin
                raise ParseError("Incorrectly-formatted bulletin: {}".format(addressee), self)

        if '{' in message:
            # Check for a message ID
            message, message_id = message.split("{")
            logger.debug("Message has message ID {}".format(message_id))

            # Message IDs must not be longer than 5 characters (C14 P71)
            if len(message_id) > 5:
                raise ParseError("Invalid message ID: {}".format(message_id), self)

            try:
                self.message = message

            except ValueError as e:
                # Message was too long
                raise ParseError("Message too long: {}".format(e))

            self.message_id = message_id
        else:
            try:
                self.message = message

            except ValueError as e:
                # Message was too long
                raise ParseError("Message too long: {}".format(e))

        return True

    @property
    def info(self) -> str:
        """Generate the information field for a message packet."""
        info = ""
        if self.addressee:
            info += self.addressee.ljust(9)
        elif self.group_bulletin_name:
            info += "BLN{}{}".format(self.bulletin_id, self.group_bulletin_name).ljust(9)
        elif self.announcement_id:
            info += "BLN{}".format(self.announcement_id).ljust(9)
        elif self.bulletin_id:
            info += "BLN{}".format(self.bulletin_id).ljust(9)
        else:
            raise GenerateError("No addressee, announcement or bulletin details", self)

        if self.message:
            info += ":{}".format(self.message)
        else:
            raise GenerateError("No message", self)

        if self.message_id:
            info += "{" + self.message_id

        return info

    def __repr__(self):
        if self.source:
            if self.group_bulletin_name:
                return "<MessagePacket: {} -> Group Bulletin {} #{}>".format(
                    self.source, self.group_bulletin_name, self.bulletin_id
                )
            elif self.bulletin_id:
                return "<MessagePacket: {} -> Bulletin #{}>".format(self.source, self.bulletin_id)
            elif self.announcement_id:
                return "<MessagePacket: {} -> Announcement {}>".format(
                    self.source, self.announcement_id
                )
            elif self.addressee:
                return "<MessagePacket: {} -> {}>".format(self.source, self.addressee)
            else:
                return "<MessagePacket: {}>".format(self.source)
        else:
            return "<MessagePacket>"



#!/usr/bin/env python

import re
import logging
import math

from enum import Enum
from geopy.point import Point

from typing import Tuple, Optional, Union

from ..exceptions import ParseError, GenerateError
from ..utils import APRSUtils
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class CompressionFix(Enum):
    """
    Enum to represent GPS fix types.

    See APRS 1.01 C9 P39.

    * `OLD` - old (last) GPS fix
    * `CURRENT` - current GPS fix
    """
    OLD = 0b00000000
    CURRENT = 0b00100000


class CompressionSource(Enum):
    """
    Enum to represent NMEA sources.

    See APRS 1.01 C9 P39.

    * `OTHER` - other
    * `GLL` - GPS GLL
    * `GGA` - GPS GGA
    * `RMC` - GPS RMC

    See https://www.gpsinformation.org/dale/nmea.htm for an explanation of the various GPS
    sentence types.
    """
    OTHER = 0b00000000
    GLL = 0b00001000
    GGA = 0b00010000
    RMC = 0b00011000


class CompressionOrigin(Enum):
    """
    Enum to represent the compression origin.

    See APRS 1.01 C9 P39.

    * `COMPRESSED` - Compressed
    * `TNC_BTEXT` - TNC BText
    * `SOFTWARE` - Software applications
    * `TBD` - TBD
    * `KPC3` - KPC3 TNC
    * `PICO` - Pico
    * `OTHER` - Other trackers
    * `DIGIPEATER` - Digipeater conversion
    """
    COMPRESSED = 0b00000000
    TNC_BTEXT = 0b00000001
    SOFTWARE = 0b00000010
    TBD = 0b00000011
    KPC3 = 0b00000100
    PICO = 0b00000101
    OTHER = 0b00000110
    DIGIPEATER = 0b00000111


class PositionPacket(GenericPacket):
    """
    Class to represent various kinds of position packets.

    This class represents packets which provide position information - including weather reports.
    """

    def __init__(self, latitude: float = 0.0, longitude: float = 0.0, ambiguity: int = 0,
                 course: int = None, speed: float = None, altitude: int = None, comment: str = None,
                 power: int = None, height: int = None, gain: int = None, directivity: int = None,
                 radio_range: int = None, strength: int = None, bearing: int = None,
                 number: float = None, df_range: int = None, quality: int = None,
                 compressed: bool = False, messaging: bool = False, data_type_id: str = "!",
                 compression_fix: CompressionFix = None,
                 compression_source: CompressionSource = None,
                 compression_origin: CompressionOrigin = None,
                 *args, **kwargs):

        super().__init__(*args, **kwargs, data_type_id=data_type_id)
        self._point = Point()
        self.latitude = latitude
        self.longitude = longitude
        self.ambiguity = ambiguity
        self.course = course
        self.speed = speed
        self.altitude = altitude
        self.comment = comment
        self.messaging = messaging

        # PHG/RNG/DFS/BRG/NRQ
        self.power = power
        self.height = height
        self.gain = gain
        self.directivity = directivity
        self.radio_range = radio_range
        self.strength = strength
        self.bearing = bearing
        self.number = number
        self.df_range = df_range
        self.quality = quality

        # Compression
        self.compressed = compressed
        self.compression_fix = compression_fix
        self.compression_source = compression_source
        self.compression_origin = compression_origin

    @property
    def point(self) -> Point:
        """Get a point representing the latitude, longitude and optionally the altitude"""
        return self._point

    @point.setter
    def point(self, value: Point):
        """Set a point representing the latitude, longitude and optionally the altitude"""
        self._point = value

    @property
    def latitude(self) -> float:
        """Get the latitude of the station"""
        return round(self._point.latitude, 6)

    @latitude.setter
    def latitude(self, value: float):
        """Set the latitude of the station"""
        self._point.latitude = value

    @property
    def longitude(self) -> float:
        """Get the longitude of the station"""
        return round(self._point.longitude, 6)

    @longitude.setter
    def longitude(self, value: float):
        """Set the longitude of the station"""
        self._point.longitude = value

    @property
    def ambiguity(self) -> int:
        """Get the ambiguity level of the packet"""
        return self._ambiguity

    @ambiguity.setter
    def ambiguity(self, value: int):
        """Set the ambiguity level of the packet"""
        self._ambiguity = value

    @property
    def course(self) -> int:
        """Get the course of the station"""
        return self._course

    @course.setter
    def course(self, value: int):
        """Set the course of the station"""
        self._course = value

    @property
    def speed(self) -> float:
        """Get the speed of the station"""
        return self._speed

    @speed.setter
    def speed(self, value: float):
        """Set the speed of the station"""
        self._speed = value

    @property
    def altitude(self) -> int:
        """Get the altitude of the station"""
        return self._point.altitude

    @altitude.setter
    def altitude(self, value: int):
        """Set the altitude of the station"""
        self._point.altitude = value

    @property
    def comment(self) -> str:
        """Get the packet's comment"""
        return self._comment

    @comment.setter
    def comment(self, value: str):
        """Set the packet's comment"""
        self._comment = value

    @property
    def power(self) -> int:
        """Get the power (in watts)"""
        return self._power

    @power.setter
    def power(self, value: int):
        """Set the power (in watts)"""
        self._power = value

    @property
    def height(self) -> int:
        """Get the antenna height above average terrain (in feet)"""
        return self._height

    @height.setter
    def height(self, value: int):
        """Set the antenna height above average terrain (in feet)"""
        self._height = value

    @property
    def gain(self) -> int:
        """Get the antenna gain (in dB)"""
        return self._gain

    @gain.setter
    def gain(self, value: int):
        """Set the antenna gain (in dB)"""
        self._gain = value

    @property
    def directivity(self) -> int:
        """Get the antenna directivity (in degrees)"""
        return self._directivity

    @directivity.setter
    def directivity(self, value: int):
        """Set the antenna directivity (in degrees)"""
        self._directivity = value

    @property
    def radio_range(self) -> int:
        """Get the radio range (in miles)"""
        return self._radio_range

    @radio_range.setter
    def radio_range(self, value: int):
        """Set the radio range (in miles)"""
        self._radio_range = value

    @property
    def strength(self) -> int:
        """Get the DF signal strength (in S-points)"""
        return self._strength

    @strength.setter
    def strength(self, value: int):
        """Set the DF signal strength (in S-points)"""
        self._strength = value

    @property
    def bearing(self) -> int:
        """Get the DF signal bearing (in degrees)"""
        return self._bearing

    @bearing.setter
    def bearing(self, value: int):
        """Set the DF signal bearing (in degrees)"""
        self._bearing = value

    @property
    def number(self) -> float:
        """Get the DF hit ratio percentage"""
        return self._number

    @number.setter
    def number(self, value: float):
        """Set the DF hit ratio percentage"""
        # TODO - Ensure this is a multiple of 12.5%
        self._number = value

    @property
    def df_range(self) -> int:
        """Get the DF range (in miles)"""
        return self._df_range

    @df_range.setter
    def df_range(self, value: int):
        """Set the DF range (in miles)"""
        self._df_range = value

    @property
    def quality(self) -> int:
        """Get the DF bearing accuracy (in degrees)"""
        return self._quality

    @quality.setter
    def quality(self, value: int):
        """Set the DF bearing accuracy (in degrees)"""
        # TODO - Ensure that this is valid
        self._quality = value

    @property
    def messaging(self) -> bool:
        """Get whether this station is message-capable or not"""
        return self._messaging

    @messaging.setter
    def messaging(self, value: bool):
        """Set whether this station is using message-capable or not"""
        if type(value) is not bool:
            raise TypeError("Value must be of type 'bool' ('{}' given)".format(type(value)))

        self._messaging = value

    @property
    def compressed(self) -> bool:
        """Get whether this packet is using compressed positions or not"""
        return self._compressed

    @compressed.setter
    def compressed(self, value: bool):
        """Set whether this packet is using compressed positions or not"""
        if type(value) is not bool:
            raise TypeError("Value must be of type 'bool' ('{}' given)".format(type(value)))

        self._compressed = value

    @property
    def compression_fix(self) -> CompressionFix:
        """Get the compression fix type."""
        return self._compression_fix

    @compression_fix.setter
    def compression_fix(self, value: Union[CompressionFix, None]):
        """Set the compression fix type."""
        if value is None or type(value) is CompressionFix:
            self._compression_fix = value
        else:
            raise TypeError("Compression fix must be of type 'CompressionFix' ('{}' given)".format(
                type(value)
            ))

    @property
    def compression_source(self) -> CompressionSource:
        """Get the compression source type."""
        return self._compression_source

    @compression_source.setter
    def compression_source(self, value: Union[CompressionSource, None]):
        """Set the compression source type."""
        if value is None or type(value) is CompressionSource:
            self._compression_source = value
        else:
            raise TypeError(
                "Compression source must be of type 'CompressionSource' ('{}' given)".format(
                    type(value)
                )
            )

    @property
    def compression_origin(self) -> CompressionOrigin:
        """Get the compression origin type."""
        return self._compression_origin

    @compression_origin.setter
    def compression_origin(self, value: Union[CompressionOrigin, None]):
        """Set the compression origin type."""
        if value is None or type(value) is CompressionOrigin:
            self._compression_origin = value
        else:
            raise TypeError(
                "Compression origin must be of type 'CompressionOrigin' ('{}' given)".format(
                    type(value)
                )
            )

    @staticmethod
    def _parse_uncompressed_position(data: str) -> Tuple[float, float, int, str, str]:
        """
        Parse uncompressed position data from a packet.

        :param str data: the information field of a packet, minus the initial data type identifier

        Given the information field of a packet with the initial data type identifier stripped, this
        will parse the latitude, longitude, position ambiguity, symbol table and symbol ID, and
        return them in a tuple.
        """
        # Decode the latitude and ambiguity
        try:
            lat, ambiguity = APRSUtils.decode_uncompressed_latitude(data[0:8])

        except ValueError as e:
            raise ParseError("Invalid latitude: {}".format(e))

        # Decode the longitude
        try:
            lng = APRSUtils.decode_uncompressed_longitude(data[9:18])

        except ValueError as e:
            raise ParseError("Invalid longitude: {}".format(e))

        logger.debug("Latitude: {} ({}) Longitude: {}".format(
            lat, ambiguity, lng
        ))

        # Parse the symbol table
        symbol_table = data[8]
        logger.debug("Symbol table: {}".format(symbol_table))

        try:
            # Parse the symbol ID
            symbol_id = data[18]
            logger.debug("Symbol: {}".format(symbol_id))
        except IndexError:
            raise ParseError("Missing symbol identifier")

        return (lat, lng, ambiguity, symbol_table, symbol_id)

    @staticmethod
    def _parse_compressed_byte(compression_byte: str) -> Tuple[CompressionFix, CompressionSource,
                                                               CompressionOrigin]:
        """
        Parse the compression byte.

        :param str compression_byte: the compression byte

        This will return a tuple of Enums denoting the compression fix, source and origin.

        For more information, see APRS 1.01 C9 P39.
        """

        t = ord(compression_byte) - 33

        # Determine the fix type
        for fix in iter(CompressionFix):
            if t | fix.value:
                break
        else:
            raise ParseError("Could not determine the compression fix type ({})".format(bin(t)))

        # Determine the source type
        for source in iter(CompressionSource):
            if t | source.value:
                break
        else:
            raise ParseError("Could not determine the compression source type ({})".format(bin(t)))

        # Determine the origin type
        for origin in iter(CompressionOrigin):
            if t | origin.value:
                break
        else:
            raise ParseError("Could not determine the compression origintype ({})".format(bin(t)))

        return [fix, source, origin]

    @classmethod
    def _parse_compressed_position(cls, data: str) -> Tuple[
        float, float, Optional[float], Optional[float], Optional[float], Optional[float],
        Optional[CompressionFix], Optional[CompressionSource], Optional[CompressionOrigin]
    ]:
        """
        Parse compressed position data from a packet.

        :param str data: the information field of a packet, minus the initial data type identifier

        Given the information field of a packet with the initial data type identifier stripped, this
        will parse the latitude and longitude, and optionally the altitude, course, speed and radio
        range.
        """

        if len(data) < 13:
            raise ValueError("Compressed position data must be at least 13 character")

        logger.debug("Compressed lat/lng is: {}".format(data))

        # Parse the compressed latitude and longitude character from the packet
        comp_lat = data[1:5]
        comp_lng = data[5:9]

        # Decode the latitude and longitude
        latitude = APRSUtils.decode_compressed_latitude(comp_lat)
        longitude = APRSUtils.decode_compressed_longitude(comp_lng)

        logger.debug("Latitude: {} Longitude: {}".format(latitude, longitude))

        # Decode the compression type, which determines what other data the packet provides.
        comp_type = "{:0>8b}".format(ord(data[12])-33)[::-1]

        altitude = None
        course = None
        speed = None
        radio_range = None

        # Check for altitude, course/speed or radio range
        if data[10] == " ":
            # If the 11th character is blank, then don't do any further parsing

            fix = None
            source = None
            origin = None

            logger.debug("No course, speed or range data")

        elif comp_type[3] == "0" and comp_type[4] == "1":
            # The altitude is obtained by subtracting 33 from the ASCII value for the c and s
            # characters, multiplying c by 91, adding s, and then raising 1.002 to the power of the
            # result.
            # See APRS 1.01 C9 P40
            c = ord(data[10]) - 33
            s = ord(data[11]) - 33
            altitude = round((1.002 ** (c * 91 + s)), 2)

            fix, source, origin = cls._parse_compressed_byte(data[12])

            logger.debug("Altitude: {}".format(altitude))

        elif 0 <= (ord(data[10])-33) <= 89:
            # The course is obtained by subtracting 33 from the ASCII value of c and then
            # multiplying it by 4
            c = ord(data[10]) - 33
            course = c * 4

            # The speed is obtained by subtracting 33 from the ASCII value of s and then raising
            # 1.08 to the power of the result, and finally subtracting 1.
            # We round the result to 1 decimal place.
            s = ord(data[11]) - 33
            speed = round((1.08 ** s) - 1, 1)

            fix, source, origin = cls._parse_compressed_byte(data[12])

            logger.debug("Course: {} Speed: {}".format(course, speed))

        elif data[10] == "{":
            # The radio range is obtained by subtracting 33 from the ASCII value of s, raising 1.08
            # to the power of the result, and finally multiplying it by 2.
            s = ord(data[11]) - 33
            radio_range = round(2 * (1.08 ** s), 2)

            fix, source, origin = cls._parse_compressed_byte(data[12])

            logger.debug("Radio range: {}".format(radio_range))
        else:
            raise ValueError("Invalid character when looking for course/speed or range: {}".format(
                data[10]
            ))

        return (latitude, longitude, altitude, course, speed, radio_range, fix, source, origin)

    @staticmethod
    def _generate_uncompressed_position(latitude: float, longitude: float, symbol_table: str,
                                        symbol_id: str, ambiguity: int = 0) -> str:
        """
        Generate uncompressed position data for a packet.

        :param float latitude: the position's latitude
        :param float longitude: the position's longitude
        :param int ambiguity: the level of ambiguity for the position
        :param str symbol_table: the symbol table reference
        :param str symbol_id: the symbol ID

        Given the latitude, longitude, symbol table, symbol ID and an optional ambiguity value, this
        will return an information field containing these values in an uncompressed format.
        """
        # Encode the latitude
        lat = APRSUtils.encode_uncompressed_latitude(latitude, ambiguity)

        # Encode the longitude
        lng = APRSUtils.encode_uncompressed_longitude(longitude, ambiguity)

        logger.debug("Latitude: {} ({}) Longitude: {}".format(
            lat, ambiguity, lng
        ))

        # Parse the symbol table
        logger.debug("Symbol table: {}".format(symbol_table))
        logger.debug("Symbol ID: {}".format(symbol_id))

        info = f"{lat}{symbol_table}{lng}{symbol_id}"

        return info

    @staticmethod
    def _generate_compressed_byte(fix: CompressionFix, source: CompressionSource,
                                  origin: CompressionOrigin) -> str:
        """
        Generate the compression byte based on the fix, source and origin values.

        :param CompressionFix fix: the fix status
        :param CompressionSource source: the NMEA source
        :param CompressionOrigin origin: the compression origin

        For more information, see APRS 1.01 C9 P39.
        """

        # Generate the byte by ORing the values
        t = fix.value | source.value | origin.value

        # Add 33, convert to an ASCII character
        compression_byte = chr(t + 33)

        return compression_byte

    @classmethod
    def _generate_compressed_position(cls, latitude: float, longitude: float, symbol_table: str,
                                      symbol_id: str, altitude: int = None, course: int = None,
                                      speed: int = None, radio_range: int = None,
                                      fix: CompressionFix = CompressionFix.OLD,
                                      source: CompressionSource = CompressionSource.OTHER,
                                      origin: CompressionOrigin = CompressionOrigin.SOFTWARE
                                      ) -> str:
        """
        Generate compressed position data for a packet.

        :param float latitude: the position's latitude
        :param float longitude: the position's longitude
        :param str symbol_table: the symbol table reference
        :param str symbol_id: the symbol ID
        :param int altitude: the altitude, in feet
        :param int course: the course, in degrees
        :param int speed: the speed, in knots
        :param int radio_range: the radio range, in miles
        :param str fix: the GPS fix type
        :param str source: the NMEA source
        :param str origin: the compression origin

        Given the latitude, longitude, symbol table, symbol ID, this will return an information
        field containing these values in an uncompressed format.

        Additional (mutually exclusive) optional values can also be added:-
        * course and speed
        * altitude
        * radio range

        The `fix`, `source` and `origin` values are used to generate the 'Compression byte' (see
        APRS 1.01 C9 P39). See :class:`CompressionFix`, :class:`CompressionSource` and
        :class:`CompressionOrigin`.

        If altitude is specified, then `source` will be overridden.
        """

        lat = APRSUtils.encode_compressed_latitude(latitude)
        lng = APRSUtils.encode_compressed_longitude(longitude)

        if course is not None and speed is not None:
            c = chr((course // 4) + 33)
            s = chr(round(math.log((speed + 1), 1.08)) + 33)
            t = cls._generate_compressed_byte(fix, source, origin)

        elif altitude is not None and altitude >= 1.0:
            # First get the exponent
            exp = round(math.log(altitude, 1.002))

            # The exponent is converted back into two numbers that fit the following equation:-
            # a * 91 + b = exp
            # Values must translate (after having 33 added to them) to a printable ASCII character.
            # This means we're limited to between 0 and 93 (33 and 126).
            a = None
            for b in range(0, 94):
                if (exp - b) % 91 == 0 and (exp - b) // 91 < 94:
                    a = (exp - b) // 91
                    break
            else:
                raise GenerateError("Could not encode altitude ({}ft)".format(altitude))

            c = chr(a + 33)
            s = chr(b + 33)

            source = CompressionSource.GGA
            t = cls._generate_compressed_byte(fix, source, origin)

        elif radio_range is not None:
            # The first character is always {
            c = "{"

            # The range is encoded as the exponent of the following:-
            # 2 * (1.08 ** exp)
            exp = round(math.log((radio_range/2), 1.08))
            s = chr(exp + 33)

            t = cls._generate_compressed_byte(fix, source, origin)

        else:
            c = " "
            s = "s"
            t = "T"

        return f"{symbol_table}{lat}{lng}{symbol_id}{c}{s}{t}"

    @staticmethod
    def _parse_data(data: str) -> Tuple[str, str, str, int, int, int, str]:
        """
        Parse additional information from the information field.

        :param str data: the information field of a packet, minus the initial data type identifier

        Position packets can have additional information in them, such as station power, antenna
        height, antenna gain, etc. These are described in APRS 1.01 C7. This will parse out the raw
        values, but not decode them.
        """

        phg = None
        rng = None
        dfs = None
        course = None
        speed = None
        altitude = None
        comment = None

        if re.match(r'^PHG[0-9]{4}', data[:7]):
            # Packet has a PHG (power, antenna height/gain/directivity) value
            phg = data[3:7]
            logger.debug("PHG is {}".format(phg))
            data = data[7:]

        elif re.match('^RNG[0-9]{4}', data[:7]):
            # Packet has an RNG (radio range) value
            rng = data[3:7]
            logger.debug("RNG is {}".format(rng))
            data = data[7:]

        elif re.match('^DFS[0-9]{4}', data[:7]):
            # Packet has a DFS (DF signal strength, antenna height/gain/directivity) value
            dfs = data[3:7]
            logger.debug("DFS is {}".format(dfs))
            data = data[7:]

        elif re.match('^[0-9]{3}/[0-9]{3}', data[:7]):
            # Packet has course and speed values
            course = int(data[:3])
            speed = int(data[4:7])
            logger.debug("Course is {}, speed is {}".format(course, speed))
            data = data[7:]

        # TODO - parse BRG/NRQ

        # Check for comment
        if len(data) > 0:

            # Check for altitude
            # As per APRS 1.01 C6 P26, altitude as /A=nnnnnn may appear anywhere in the comment
            has_altitude = re.match('.*/A=([0-9]{6}).*', data)
            if has_altitude:
                # TODO - fix altitude format
                altitude = int(has_altitude.groups()[0])
                logger.debug("Altitude is {} ft".format(altitude))

                # Strip out the altitude from the comment
                data = re.sub(r'/A=[0-9]{6}', "", data)

            # Set the comment as the remainder of the information field
            comment = data
            logger.debug("Comment is {}".format(comment))

        return (phg, rng, dfs, course, speed, altitude, comment)

    @staticmethod
    def _generate_data(phg: str = None, rng: str = None, dfs: str = None, course: int = None,
                       speed: int = None, altitude: int = None, comment: str = None) -> str:
        """
        Generate additional information for the information field.

        :param str phg: a PHG value
        :param str rng: an RNG value
        :param str dfs: a DFS value
        :param int course: course, in degrees
        :param int speed: speed, in knots
        :param int altitude: altitude, in feet
        :param str comment: a comment

        Position packets can have additional information in them, such as station power, antenna
        height, antenna gain, etc. These are described in APRS 1.01 C7. This will generate an
        information field based on these values.

        Note that PHG, RNG, DFS and course/speed are mutually exclusive. Altitude is specified as
        part of the comment and so can coexist with any of the other values.
        """

        data = ""
        if phg:
            data += phg
        elif rng:
            data += rng
        elif dfs:
            data += dfs
        elif course and speed:
            data += "{}/{}".format(
                str(course).zfill(3), str(speed).zfill(3)
            )

        if altitude:
            data += "/A={}".format(
                str(altitude).zfill(6)
            )

        if comment:
            data += comment

        return data

    def _parse(self) -> bool:
        """
        Parse a position packet.

        There are a number of different position packet types - with or without a timestamp, and
        with or without messaging capability. The data type ID is used to distinguish between them.

        The position data itself can either be compressed or uncompressed, regardless of the data
        type ID.

        The parsed and decoded values are stored in the current object.
        """

        # First, check if this packet has a '!' at an offset position
        # This is allowed as per APRS 1.01 C5 P18
        if hasattr(self, '_offset'):
            # Packets with the '!' offset do not have a timestamp or messaging capabilities
            # Chop everything off the info field before the '!'
            self._info = self._info[self._offset:]

        elif self.data_type_id == '!':
            # Packet has no timestamp, station has no messaging capability
            self.timestamp = None
            self.messaging = False

        elif self.data_type_id == '/':
            # Packet has timestamp, station has no messaging capability
            self.messaging = False

            # Parse timestamp
            (self.timestamp, self.timestamp_type) = APRSUtils.decode_timestamp(self._info[0:8])

        elif self.data_type_id == '=':
            # Packet has no timestamp, station has messaging capability
            self.timestamp = None
            self.messaging = True

        elif self.data_type_id == '@':
            # Packet has timestamp, station has messaging capability
            self.messaging = True

            # Parse timestamp
            (self.timestamp, self.timestamp_type) = APRSUtils.decode_timestamp(self._info[0:8])

        else:
            # This isn't a position packet
            raise ParseError("Unknown position data type: {}".format(self.data_type_id))

        if self.timestamp is None:
            data = self._info
        else:
            data = self._info[7:]

        # Check to see if the position data is compressed or uncompressed
        if re.match(r'[0-9\s]{4}\.[0-9\s]{2}[NS].[0-9\s]{5}\.[0-9\s]{2}[EW]', data):
            # Parse the uncompressed position values from the information field
            (self.latitude, self.longitude, self.ambiguity, self.symbol_table, self.symbol_id
             ) = self._parse_uncompressed_position(data)

            # Ensure compressed is set to False
            self.compressed = False

            if len(data) > 19:
                # This packet has additional data in the information field, so attempt to parse it
                (phg, radio_range, dfs, self.course, self.speed, self.altitude,
                 comment) = self._parse_data(data[19:])

                if self.symbol_table == "/" and self.symbol_id == "\\":
                    # If the symbol table is /, and the symbol ID is \, it implies a DF report
                    # 26th and 30th characters should be /
                    logger.debug("Symbol table and symbol indicates a DF report")

                    if len(comment) < 8:
                        # Packets with DF information must be at least 8 characters long
                        raise ParseError("Missing DF values")

                    if comment[0] != "/" or comment[4] != "/":
                        # Packets with DF information must also include the bearing and NRQ values
                        # See APRS 1.01 C7 P30
                        raise ParseError(
                            "Invalid DF values (character in position 0 and 4 should be '/'"
                        )

                    # Extract the bearing
                    self.bearing = int(comment[1:4])
                    logger.debug(f"DF bearing is {self.bearing} degrees")

                    # Decode the NRQ value
                    (self.number, self.df_range, self.quality) = APRSUtils.decode_nrq(comment[5:8])

                    # Strip the bearing/NRQ value from the comment
                    self.comment = comment[8:]

                elif self.symbol_table in ["/", "\\"] and self.symbol_id == "_":
                    # / or \, and _ for the symbol table and symbol implies a weather report
                    # TODO - Implementation
                    logger.debug("Symbol table and symbol indicates a weather report")

                elif phg:
                    # Decode the power, height, gain and directivity values
                    (self.power, self.height, self.gain, self.directivity) = \
                        APRSUtils.decode_phg(phg)

                    # The PHG value has already been stripped from the comment
                    self.comment = comment

                elif radio_range:
                    # The radio range is specified as 4 digits, which denote the range in miles
                    self.radio_range = int(radio_range)
                    logger.debug(f"Radio range is {radio_range} miles")

                    # The PHG value has already been stripped from the comment
                    self.comment = comment

                elif dfs:
                    # Decode the signal strength, height, gain and directivity values
                    (self.strength, self.height, self.gain, self.directivity) = \
                        APRSUtils.decode_dfs(dfs)

                    # The PHG value has already been stripped from the comment
                    self.comment = comment

                else:
                    # No additional data found
                    self.comment = comment

        else:
            # Parse the compressed position values from the information field

            # Get the compressed position
            compressed_position = data[0:13]

            try:
                (self.latitude, self.longitude, self.altitude, self.course, self.speed,
                self.radio_range, self.compression_fix, self.compression_source,
                self.compression_origin) = self._parse_compressed_position(compressed_position)

            except Exception as e:
                # TODO Catch specific errors (for example, OverflowError)
                raise ParseError("Couldn't parse compressed position: {}".format(e))

            # Ensure compressed is set to True
            self.compressed = True

            # Parse the symbol table and symbol ID
            self.symbol_table = data[0]
            self.symbol_id = data[9]

            # TODO - parse altitude information

            self.comment = data[13:]
            logger.debug("Comment is {}".format(self.comment))

        # If we get this far, then we've parsed the packet
        return True

    @property
    def info(self) -> str:
        """Generate the information field for a position packet."""
        info = ""

        # Ensure we have a latitude and a longitude
        if self.latitude is None:
            raise GenerateError("Missing latitude")
        elif self.longitude is None:
            raise GenerateError("Missing longitude")

        # Ensure we have a symbol table and symbol ID
        if self.symbol_table is None:
            raise GenerateError("Missing symbol table")
        elif self.symbol_id is None:
            raise GenerateError("Missing symbol ID")

        # Set data type ID
        if self.timestamp is None:
            if self.messaging is False:
                self.data_type_id = "!"
            else:
                self.data_type_id = "="
        else:
            if self.messaging is False:
                self.data_type_id = "/"
            else:
                self.data_type_id = "@"

            # Set the timestamp
            info += APRSUtils.encode_timestamp(self.timestamp, self.timestamp_type)

        if self.compressed:
            # Add the position in a compressed format
            info += self._generate_compressed_position(
                self.latitude, self.longitude, self.symbol_table, self.symbol_id, self.altitude,
                self.course, self.speed, self.radio_range, self.compression_fix,
                self.compression_source, self.compression_origin)

            # PHG, etc is not supported for compressed formats (see APRS 1.01 C9 P36)
            if self.comment:
                info += self.comment

        else:
            # Add the position in an uncompressed format
            # TODO: handle BRG/NRQ
            info += self._generate_uncompressed_position(
                self.latitude, self.longitude, self.symbol_table, self.symbol_id, self.ambiguity
            )

            # Handle PHG
            if self.power is not None and self.height is not None and self.gain is not None \
                    and self.directivity is not None:
                phg = APRSUtils.encode_phg(self.power, self.height, self.gain, self.directivity)
                info += "PHG{}".format(
                    self._generate_data(phg=phg, altitude=self.altitude, comment=self.comment)
                )

            # Handle DFS
            elif self.strength is not None and self.height is not None and self.gain is not None \
                    and self.directivity is not None:
                dfs = APRSUtils.encode_dfs(self.strength, self.height, self.gain, self.directivity)
                info += "DFS{}".format(
                    self._generate_data(dfs=dfs, altitude=self.altitude, comment=self.comment)
                )

            # Handle course/speed
            elif self.course is not None and self.speed is not None:
                info += "{}/{}".format(
                    str(self.course).zfill(3),
                    str(self.speed).zfill(3)
                )
                info += self._generate_data(altitude=self.altitude, comment=self.comment)

            # Handle RNG
            elif self.radio_range is not None:
                info += "RNG{}".format(
                    str(self.radio_range).zfill(4)
                )
                info += self._generate_data(altitude=self.altitude, comment=self.comment)

            else:
                info += self._generate_data(altitude=self.altitude, comment=self.comment)

        return info

    def __repr__(self):
        if self.source:
            return "<PositionPacket: {}>".format(self.source)
        else:
            return "<PositionPacket>"



#!/usr/bin/env python

import re
import logging
import math
from datetime import datetime, timedelta, UTC
from typing import Union, Tuple, Optional
from .exceptions import ParseError

# Set up logging
logger = logging.getLogger(__name__)


class APRSUtils:
    """
    Useful generic functions for APRS parsing and decoding.

    This class provides functions for parsing and decoding different kinds of APRS packets.
    Packet type-specific functions are defined under the different :class:`APRSPacket` subclasses.
    """
    @staticmethod
    def _get_utc():
        return datetime.now(UTC)

    @staticmethod
    def decode_uncompressed_latitude(latitude: str) -> Tuple[Union[int, float], int]:
        """
        Convert an uncompressed latitude string to a latitude and an ambiguity
        value.

        :param str latitude: an uncompressed latitude, in the form ``DDMM.HHC``

        Uncompressed latitudes have the format DDMM.HHC, where:-
         * DD are the degrees
         * MM are the minutes
         * HH are the hundredths of minutes
         * C is either N (for the northern hemisphere) or S (southern)

        MM and HH can be replaced with spaces (" ") to provide positional
        ambiguity (C6 P24).

        See also APRS 1.01 C6 P23.
        """
        logger.debug("Input latitude: {}".format(latitude))

        # Regex match to catch any obviously-invalid latitudes
        if not re.match(r'^[0-9]{2}[\s0-9]{2}\.[\s0-9]{2}[NS]$', latitude):
            raise ValueError("Invalid latitude: {}".format(latitude))

        # Determine the level of ambiguity, and replace the spaces with zeroes
        # See C6 P24
        ambiguity = latitude.count(' ')
        latitude = latitude.replace(' ', '0')
        logger.debug("Ambiguity: {}".format(ambiguity))

        try:
            # Extract the number of degrees
            degrees = int(latitude[:2])
            if degrees > 90:
                raise ValueError("Invalid degrees: {}".format(degrees))

            logger.debug("Degrees: {}".format(degrees))

            # Extract the number of minutes, convert it to a fraction of a degree,
            # and round it to 6 decimal places
            minutes = round(float(latitude[2:-1])/60, 6)
            logger.debug("Minutes: {}".format(minutes))

            # Extract the north/south
            direction = latitude[-1]
            logger.debug("Direction: {}".format(direction))

        except Exception as e:
            raise ParseError("Couldn't parse latitude {}: {}".format(
                latitude, e
            ))

        # Add the degrees and minutes to give the latitude
        lat = degrees + minutes

        # If the latitude is south of the equator, make it negative
        if direction == "S":
            lat *= -1

        logger.debug("Output latitude: {}, ambiguity: {}".format(
            lat, ambiguity
        ))
        return (lat, ambiguity)

    @staticmethod
    def encode_uncompressed_latitude(latitude: Union[float, int], ambiguity: int = 0) -> str:
        """
        Encode a latitude into an uncompressed latitude format.

        :param float/int latitude: a latitude
        :param int ambiguity: an optional ambiguity level

        For more information see :func:`decode_uncompressed_latitude`
        """

        # The latitude must be a float
        if type(latitude) is not float:
            if type(latitude) is int:
                latitude = float(latitude)
            else:
                raise TypeError("Latitude must be an float ({} given)".format(type(latitude)))

        # The latitude must be between -90 and 90 inclusive
        if not -90 <= latitude <= 90:
            raise ValueError("Latitude must be between -90 and 90 inclusive ({} given)".format(
                latitude
            ))

        # The ambiguity must be an int
        if ambiguity and type(ambiguity) is not int:
            raise TypeError("Ambiguity must be an int ({} given)".format(type(ambiguity)))

        # The ambiguity must be between 0 and 4 inclusive
        if ambiguity < 0 or ambiguity > 4:
            raise ValueError("Ambiguity must be between 0 and 4 inclusive ({} given)".format(
                ambiguity
            ))

        # Determine if the latitude is north or south
        if latitude < 0:
            # Direction is south
            direction = "S"
            latitude *= -1
        else:
            direction = "N"

        # Get the degrees of latitude
        degrees = math.floor(latitude)

        # Get the minutes of latitude
        minutes = "{:0.2f}".format(round((latitude - degrees) * 60, 2)).zfill(5)

        # Zero-pad the degrees
        degrees_padded = str(degrees).zfill(2)

        # Apply ambiguity
        if ambiguity == 0:
            lat = "{}{}".format(degrees_padded, minutes)
        elif ambiguity == 1:
            lat = "{}{} ".format(degrees_padded, minutes[:-1])
        elif ambiguity == 2:
            lat = "{}{}  ".format(degrees_padded, minutes[:-2])
        elif ambiguity == 3:
            lat = "{}{} .  ".format(degrees_padded, minutes[:-4])
        elif ambiguity == 4:
            lat = "{}  .  ".format(degrees_padded)

        return f"{lat}{direction}"

    @staticmethod
    def encode_uncompressed_longitude(longitude: float, ambiguity: int = 0) -> str:
        """
        Encode a longitude into an uncompressed longitude format.

        :param float longitude: a longitude
        :param int ambiguity: an optional ambiguity level

        For more information see :func:`decode_uncompressed_latitude`
        """

        # The longitude must be a float
        if type(longitude) is not float:
            if type(longitude) is int:
                longitude = float(longitude)
            else:
                raise TypeError("Longitude must be a float ({} given)".format(type(longitude)))

        # The longitude must be between -180 and 180 inclusive
        if not -180 <= longitude <= 180:
            raise ValueError("Longitude must be between -180 and 180 inclusive ({} given)".format(
                longitude
            ))

        # The ambiguity must be an int
        if ambiguity and type(ambiguity) is not int:
            raise TypeError("Ambiguity must be an int ({} given)".format(type(ambiguity)))

        # The ambiguity must be between 0 and 4 inclusive
        if ambiguity < 0 or ambiguity > 4:
            raise ValueError("Ambiguity must be between 0 and 4 inclusive ({} given)".format(
                ambiguity
            ))

        # Determine if the longitude is east or west
        if longitude < 0:
            # Direction is south
            direction = "W"
            longitude *= -1
        else:
            direction = "E"

        # Get the degrees of longitude
        degrees = math.floor(longitude)

        # Get the minutes of longitude
        minutes = "{:0.2f}".format(round((longitude - degrees) * 60, 2)).zfill(5)

        # Zero-pad the degrees
        degrees_padded = str(degrees).zfill(3)

        # Apply ambiguity
        if ambiguity == 0:
            lng = "{}{}".format(degrees_padded, minutes)
        elif ambiguity == 1:
            lng = "{}{} ".format(degrees_padded, minutes[:-1])
        elif ambiguity == 2:
            lng = "{}{}  ".format(degrees_padded, minutes[:-2])
        elif ambiguity == 3:
            lng = "{}{} .  ".format(degrees_padded, minutes[:-4])
        elif ambiguity == 4:
            lng = "{}  .  ".format(degrees_padded)

        return f"{lng}{direction}"

    @staticmethod
    def decode_uncompressed_longitude(longitude: str, ambiguity: int = 0) -> float:
        """
        Convert an uncompressed longitude string to a longitude value, with an
        optional ambiguity level applied.

        :param str longitude: the longitude, in the format ``DDDMM.HHC``
        :param int ambiguity: the level of ambiguity, between 1 and 4

        Uncompressed longitudes have the format DDDMM.HHC, where:-
         * DD are the degrees
         * MM are the minutes
         * HH are the hundreths of minutes
         * C is either W (for west of the meridian) or E (for east)

        Positional ambiguity is handled by the latitude, and so should be
        honoured regardless of the precision of the longitude given (as per C6 P24).
        """
        logger.debug("Input longitude: {}, ambiguity: {}".format(
            longitude, ambiguity
        ))

        # Regex match to catch any obviously-invalid longitudes
        if not re.match(r'^[0-1][0-9]{2}[\s0-9]{2}\.[\s0-9]{2}[EW]$', longitude):
            raise ValueError("Invalid longitude: {}".format(longitude))

        # If the ambiguity value is more than 0, replace longitude digits with
        # spaces as required.
        if 0 <= ambiguity <= 4:
            longitude = longitude.replace(' ', '0')
        else:
            raise ValueError("Invalid ambiguity level: {} (maximum is 4)".format(
                ambiguity
            ))

        # Extract the number of degrees
        degrees = int(longitude[:3])
        if degrees > 180:
            raise ValueError("Invalid degrees: {}".format(degrees))

        logger.debug("Degrees: {}".format(degrees))

        # Since we don't want to replace the '.', if the ambiguity is more than
        # 2, increment it by 1.
        if ambiguity > 2:
            ambiguity_level = ambiguity + 1
        else:
            ambiguity_level = ambiguity

        # Extract the number of minutes, and round it to 6 decimal places
        if ambiguity == 1:
            mins = longitude[3:-2] + "0"
        elif ambiguity == 2:
            mins = longitude[3:-3] + "00"
        elif ambiguity == 3:
            mins = longitude[3:-5] + "0.00"
        elif ambiguity == 4:
            mins = longitude[3:-6] + "00.00"
        else:
            mins = longitude[3:-1]

        minutes = round(float(mins)/60, 6)
        logger.debug("Ambiguity level: {}".format(ambiguity_level))
        logger.debug("Minutes: {}".format(minutes))

        # Extract the east/west
        direction = longitude[-1]

        logger.debug("Direction: {}".format(direction))

        # Add the degrees and minutes to give the longitude
        lng = degrees + minutes

        # If the longitude is west of the meridian, make it negative
        if direction == "W":
            lng *= -1

        logger.debug("Output longitude: {}".format(lng))
        return lng

    @staticmethod
    def decode_compressed_latitude(latitude: str) -> float:
        """
        Convert a compressed latitude string to a latitude value.

        :param str latitude: a latitude in a compressed format

        Compressed latitudes have the format YYYY, where all values are base-91
        printable ASCII characters.

        See also APRS 1.01 C9 P38.
        """
        logger.debug("Input compressed latitude: {}".format(latitude))

        # The compressed latitude string must be 4 characters
        if len(latitude) != 4:
            raise ValueError("Input compressed latitude must be 4 characters ({} given)".format(
                len(latitude)
            ))

        # As per APRS 1.01, if the compressed latitude is y1y2y3y4, the latitude
        # can be determined with:-
        # 90 - ((y1-33) x 913 + (y2-33) x 912 + (y3-33) x 91 + y4-33) / 380926
        lat = 90 - (
            (ord(latitude[0])-33) * 91**3 +
            (ord(latitude[1])-33) * 91**2 +
            (ord(latitude[2])-33) * 91 +
            (ord(latitude[3])-33)
        ) / 380926
        lat = round(lat, 6)

        if lat > 90 or lat < -91:
            raise ParseError("Invalid compressed latitude (greater than 90 or less than -90)")

        # Return latitude
        logger.debug("Output latitude: {}".format(lat))
        return lat

    @staticmethod
    def decode_compressed_longitude(longitude: str) -> float:
        """
        Convert a compressed longitude string to a latitude value.

        :param str longitude: a longitude in compressed format

        Compressed longitude have the format XXXX, where all values are base-91
        printable ASCII characters

        See also APRS 1.01 C9 P38
        """
        logger.debug("Input compressed longitude: {}".format(longitude))

        # The compressed longitude must be 4 characters
        if len(longitude) != 4:
            raise ValueError("Input compressed longitude must be 4 characters ({} given)".format(
                len(longitude)
            ))

        # If the compressed longitude is x1x2x3x4, the longitude can be determined
        # with:-
        # -180 + ((x1-33) x 913 + (x2-33) x 912  + (x3-33) x 91 + x4-33) / 190463
        lng = -180 + (
            (ord(longitude[0])-33) * 91**3 +
            (ord(longitude[1])-33) * 91**2 +
            (ord(longitude[2])-33) * 91 +
            (ord(longitude[3])-33)
        ) / 190463
        lng = round(lng, 6)

        if lng > 180 or lng < -180:
            raise ParseError("Invalid compressed longitude (greater than 180 or less than -180)")

        # Return longitude
        logger.debug("Output longitude: {}".format(lng))
        return lng

    @staticmethod
    def encode_compressed_latitude(latitude: Union[float, int]) -> str:
        """
        Convert a latitude to a compressed latitude value.

        :param float/int latitude: a latitude

        Compressed latitudes have the format YYYY, where all values are base-91
        printable ASCII characters.

        See also APRS 1.01 C9 P38.
        """
        logger.debug("Input latitude: {}".format(latitude))

        # Ensure float or int, and between -90 and 90
        if type(latitude) is not float and type(latitude) is not int:
            raise TypeError("Latitude must be a 'float' or 'int' ({} given)".format(type(latitude)))

        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 ({} given)".format(latitude))

        total = 380926 * (90 - latitude)
        (a, remainder) = divmod(total, 91 ** 3)
        (b, remainder) = divmod(remainder, 91 ** 2)
        (c, d) = divmod(remainder, 91 ** 1)

        lat = "{}{}{}{}".format(
            chr(int(a)+33),
            chr(int(b)+33),
            chr(int(c)+33),
            chr(int(d)+33)
        )

        # Return latitude
        logger.debug("Output latitude: {}".format(lat))
        return lat

    @staticmethod
    def encode_compressed_longitude(longitude: Union[float, int]) -> str:
        """
        Convert a longitude to a compressed latitude value.

        :param float/int longitude: a longitude

        Compressed longitude have the format XXXX, where all values are base-91
        printable ASCII characters

        See also APRS 1.01 C9 P38
        """
        logger.debug("Input longitude: {}".format(longitude))

        # Ensure float or int, and between -180 and 180
        if type(longitude) is not float and type(longitude) is not int:
            raise TypeError("Longitude must be a 'float' or 'int' ({} given)".format(
                type(longitude)
            ))

        if longitude < -180 or longitude > 180:
            raise ValueError("Longitude must be between -180 and 180 ({} given)".format(longitude))

        total = 190463 * (180 + longitude)
        (a, remainder) = divmod(total, 91 ** 3)
        (b, remainder) = divmod(remainder, 91 ** 2)
        (c, d) = divmod(remainder, 91 ** 1)

        lng = "{}{}{}{}".format(
            chr(int(a)+33),
            chr(int(b)+33),
            chr(int(c)+33),
            chr(int(d)+33)
        )

        # Return longitude
        logger.debug("Output longitude: {}".format(lng))
        return lng

    @staticmethod
    def decode_timestamp(raw_timestamp: str) -> datetime:
        """
        Decode a timestamp.

        :param str raw_timestamp: a string representing a timestamp

        Timestamps can take a number of different forms:-
         * Zulu, identified with a trailing 'z', which refers to zulu time
         * Local, identified with a trailing '/', which has no timezone information
         * A hour/minute/second timestamp without any date information
        """

        logger.debug("Raw timestamp is {}".format(raw_timestamp))
        ts = re.match(r'^(\d{6})(.)', raw_timestamp)
        if ts:
            timestamp, timestamp_type = ts.groups()

            # Parse the timestamp type
            # APRS 1.01 specifies 'z' for zulu time, but some clients use 'Z'
            if timestamp_type == 'z' or timestamp_type == 'Z':
                timestamp_type = 'zulu'
                logger.debug("Timestamp is zulu time")
            elif timestamp_type == '/':
                timestamp_type = 'local'
                logger.debug("Timestamp is local time")
            elif timestamp_type == 'h':
                timestamp_type = 'hms'
                logger.debug("Timestamp is hhmmss time")
            else:
                # This is against spec, but as usual with APRS a lot of clients violate this - so
                # assume they're zulu time
                logger.warning("{} is an invalid timestamp type, assuming zulu".format(timestamp_type))
                timestamp_type = 'zulu'

            # Get the current UTC ('zulu') time for comparison. Since timestamps in HHMMSS format
            # have no date information, we need to determine if the timestamp is for today or
            # yesterday
            utc = APRSUtils._get_utc()

            if timestamp_type == 'hms':
                # HHMMSS format

                # Parse out the hour, minute and second
                hour = int(timestamp[0:2])
                minute = int(timestamp[2:4])
                second = int(timestamp[4:6])

                try:
                    # Generate a datetime object
                    ts = datetime(
                        year=utc.year,
                        month=utc.month,
                        day=utc.day,
                        hour=hour,
                        minute=minute,
                        second=second,
                        tzinfo=UTC
                    )
                except ValueError as e:
                    logger.error("Error parsing timestamp '{}': {}".format(raw_timestamp, e))
                    raise ParseError("Error parsing timestamp '{}': {}".format(raw_timestamp, e))

                # Check it's not in the future
                if ts > utc:
                    # The timestamp is in the future, so go back a day
                    ts -= timedelta(days=1)

                # Convert to seconds
                timestamp = int(ts.timestamp())
                logger.debug("Timestamp is {}".format(ts.strftime("%Y%m%d%H%M%S")))

            elif timestamp_type == 'zulu' or timestamp_type == 'local':
                if timestamp_type == 'local':
                    # Assume local time is zulu time
                    # NOTE: this is against the spec, but without knowing the local
                    # timezone, it's impossible to work out. APRS 1.01 C6 P22 states "It is
                    # recommended that future APRS implementations only transmit zulu format on the
                    # air", so hopefully this shouldn't be a problem in reality.
                    logger.info("Local time specified in timestamp, assuming UTC.")

                # Sometimes, 000000 is used to indicate no timestamp is available
                if timestamp[0:6] == "000000":
                    logger.warning("Timestamp specified but is set to all zeroes.")
                    return None, timestamp_type

                # DDHHMM
                day = int(timestamp[0:2])
                hour = int(timestamp[2:4])
                minute = int(timestamp[4:6])

                # TODO - handle broken timestamps a bit nicer
                try:
                    ts = datetime(
                        year=utc.year,
                        month=utc.month,
                        day=day,
                        hour=hour,
                        minute=minute,
                        second=0,
                        tzinfo=UTC
                    )
                except ValueError as e:
                    logger.warning("Error parsing timestamp '{}': {}".format(timestamp, e))
                    raise ParseError("Error parsing timestamp '{}': {}".format(timestamp, e))

                # Check it's not in the future
                if ts > utc:
                    logger.debug("Timestamp day in previous month.")
                    # The time is in the future, so go back a month
                    # timedelta doesn't support subtracting months, so here be
                    # dirty hacks.
                    if 1 < ts.month <= 12:
                        month = ts.month - 1
                        ts = datetime(
                            year=utc.year,
                            month=month,
                            day=day,
                            hour=hour,
                            minute=minute,
                            second=0,
                            tzinfo=UTC
                        )
                    elif ts.month == 1:
                        year = ts.year - 1
                        month = 12
                        ts = datetime(
                            year=year,
                            month=month,
                            day=day,
                            hour=hour,
                            minute=minute,
                            second=0,
                            tzinfo=UTC
                        )

                # Convert to seconds
                logger.debug("Timestamp is {}".format(ts.strftime("%Y%m%d%H%M%S")))

            return ts, timestamp_type

        else:
            raise ParseError("No valid timestamp found")

    @staticmethod
    def encode_timestamp(timestamp: datetime, timestamp_type: str = "zulu") -> str:
        """
        Encode a timestamp.

        :param datetime timestamp: a timestamp
        :param str timestamp_type: the timestamp type (`zulu`, `hms` or `local`)
        """
        if type(timestamp) is not datetime:
            raise TypeError("Timestamp must be of type 'datetime' ({} given)".format(
                type(timestamp)
            ))

        if timestamp_type == "zulu":
            return "{}{}{}z".format(
                str(timestamp.day).zfill(2),
                str(timestamp.hour).zfill(2),
                str(timestamp.minute).zfill(2)
            )
        elif timestamp_type == "local":
            return "{}{}{}/".format(
                str(timestamp.day).zfill(2),
                str(timestamp.hour).zfill(2),
                str(timestamp.minute).zfill(2)
            )
        elif timestamp_type == "hms":
            return "{}{}{}h".format(
                str(timestamp.hour).zfill(2),
                str(timestamp.minute).zfill(2),
                str(timestamp.second).zfill(2)
            )
        else:
            raise ValueError(
                "Timestamp type must be one of 'zulu', 'hms' or 'local' ({} given)".format(
                    timestamp_type
                )
            )

    @staticmethod
    def decode_phg(phg: str) -> Tuple[int, int, int, Optional[int]]:
        """
        Decode a PHG (Power, Effective Antenna Height/Gain/Directivity) value and return the
        individual values.

        :param str phg: a PHG value, minus the initial ``PHG`` identifier

        The PHG extension provides a way of specifying approximate values for:-
         * Power (in watts)
         * Height above average local terrain (in feet)
         * Antenna gain (in dB)
         * Directivity (in degrees)

        PHG values are 4 characters long, and each digit is responsible for one of the above values.
        As per the APRS spec, the height value can be any ASCII character from 0 upwards.

        See APRS 1.01 C7 P28
        """

        # Ensure this is a valid PHG value
        if not re.match(r'^[0-9]\w[0-9]{2}$', phg):
            raise ValueError(f"Invalid PHG value: {phg}")

        # Decode the power value
        # To get this, we square the value
        power = int(phg[0])**2
        logger.debug(f"Power is {power} watts")

        # Decode the height value
        # To get this, we raise 2 to the power of the ASCII value of the character, minus 48, then
        # times it by 10
        height = (2 ** (ord(phg[1])-48)) * 10
        logger.debug(f"Height is {height} feet")

        # Decode the gain value
        # The gain is just the digit, in dB
        gain = int(phg[2])
        logger.debug(f"Gain is {gain}dB")

        # Decode the directivity
        # The directivity is 360 divided by 8, times the directivity value
        # A value of 0 implies an omnidirectional antenna
        if int(phg[3]) == 0:
            directivity = None
            logger.debug(f"Directivity is omnidirectional")
        else:
            directivity = int((360/8) * int(phg[3]))
            logger.debug(f"Directivity is {directivity} degrees")

        return (power, height, gain, directivity)

    @staticmethod
    def encode_phg(power: int, height: int, gain: int, directivity: Union[int, str]) -> str:
        """
        Encode a PHG (Power, Effective Antenna Height/Gain/Directivity) value from individual
        values.

        :param int power: the power, in watts
        :param int height: the antenna height, in feet
        :param int gain: the antenna gain, in dB
        :param int/str directivity: the antenna directivity, in degrees

        For more information, see :func:`decode_phg`.
        """

        # Validate the given values
        # Power must be a value between 0 to 9 squared
        if type(power) is not int:
            raise TypeError("Power must be of type 'int' ({} given)".format(type(power)))
        elif power not in [v ** 2 for v in range(0, 10)]:
            raise ValueError("Power must be one of {} ({} given)".format(
                [v ** 2 for v in range(0, 10)], power
            ))
        else:
            p = int(math.sqrt(power))
            logger.debug(f"Encoded power {power} is {p}")

        # Antenna height must be a value that is 2 raised to the power of the ASCII value higher
        # than or equal to '0' minus 48, then multiplied by 10
        if type(height) is not int:
            raise TypeError("Height must be of type 'int' ({} given)".format(type(height)))
        elif height not in [(2 ** (v-48))*10 for v in range(48, 127)]:
            raise ValueError("Height must be one of {} ({} given)".format(
                [(2 ** (v-48))*10 for v in range(48, 127)], height
            ))
        else:
            h = chr(int((math.log((height/10), 2)+48)))

        # Antenna gain must be a number between 0 and 9 inclusive
        if type(gain) is not int:
            raise TypeError("Gain must be of type 'int' ({} given)".format(type(gain)))
        elif not 0 <= gain <= 9:
            raise ValueError("Gain must be between 0 and 9 inclusive ({} given)".format(
                gain
            ))
        else:
            g = gain

        # Directivity must be a multiple of 45
        if directivity is None:
            d = 0
        elif type(directivity) is int:
            if directivity % (360/8) != 0.0:
                raise ValueError("Directivity must be a multiple of 45 ({} given)".format(
                    directivity
                ))
            else:
                d = int(directivity / 45)
        else:
            raise TypeError(
                "Directivity must either be of type 'int' or None ({} given)".format(
                    type(directivity)
                ))

        # Return the encoded PHG value
        return f"{p}{h}{g}{d}"

    @staticmethod
    def decode_dfs(dfs: str) -> Tuple[int, int, int, Optional[int]]:
        """
        Decode a DFS (Omni-DF Signal Strength) value and return the individual values.

        :param str dfs: a DFS value, minus the initial ``DFS`` identifier

        The DFS extension provides a way of specifying approximate values for:-
         * Received signal strength (in S-points)
         * Antenna height above average terrain (in feet)
         * Antenna gain (in dB)
         * Directivity (in degrees)

        Like PHG values, DFS values are 4 characters long, and each digit is responsible for one of
        the above values. The APRS spec does not specify it, but it is assumed that - like the PHG
        value - the antenna height can be any ASCII value from 0 upwards.

        See APRS 1.01 C7 P29
        """

        # Ensure this is a valid DFS value
        if not re.match(r'^[0-9]\w[0-9]{2}$', dfs):
            raise ValueError(f"Invalid DFS value: {dfs}")

        # Decode the strength value
        # This just the digit, in dB
        strength = int(dfs[0])
        logger.debug(f"Strength is S{strength}")

        # Decode the height value
        # To get this, we raise 2 to the power of the ASCII value of the character, minus 48, then
        # times it by 10
        height = (2 ** (ord(dfs[1])-48)) * 10
        logger.debug(f"Height is {height} feet")

        # Decode the gain value
        # The gain is just the digit, in dB
        gain = int(dfs[2])
        logger.debug(f"Gain is {gain}dB")

        # Decode the directivity
        # The directivity is 360 divided by 8, times the directivity value
        # A value of 0 implies an omnidirectional antenna
        if int(dfs[3]) == 0:
            directivity = None
            logger.debug(f"Directivity is omnidirectional")
        else:
            directivity = (360/8) * int(dfs[3])
            logger.debug(f"Directivity is {directivity} degrees")

        return (strength, height, gain, directivity)

    @staticmethod
    def encode_dfs(strength: int, height: int, gain: int, directivity: int = None) -> str:
        """
        Encode a DFS (Omni-DF Signal Strength) value from individual values.

        :param int strength: the received signal strength, in S-points
        :param int height: the antenna height, in feet
        :param int gain: the antenna gain, in dB
        :param int/str directivity: the antenna directivity, in degrees, or ``omni``

        For more information, see :func:`decode_dfs`.
        """

        # Signal strength must be a digit between 0 and 9 inclusive

        if type(strength) is not int:
            raise TypeError("Signal strength must be of type 'int' ({} given)".format(
                type(strength)
            ))
        elif not 0 <= strength <= 9:
            raise ValueError(
                "Signal strength must be between 0 and 9 inclusive ({} given)".format(
                    strength
                ))
        else:
            s = strength

        # Antenna height must be a value that is 2 raised to the power of the ASCII value higher
        # than or equal to '0' minus 48, then multiplied by 10
        if type(height) is not int:
            raise TypeError("Height must be of type 'int' ({} given)".format(type(height)))
        elif height not in [(2 ** (v-48))*10 for v in range(48, 127)]:
            raise ValueError("Height must be one of {} ({} given)".format(
                [(2 ** (v-48))*10 for v in range(48, 127)], height
            ))
        else:
            h = chr(int((math.log((height/10), 2)+48)))

        # Antenna gain must be a number between 0 and 9 inclusive
        if type(gain) is not int:
            raise TypeError("Gain must be of type 'int' ({} given)".format(type(gain)))
        if not 0 <= gain <= 9:
            raise ValueError("Gain must be between 0 and 9 inclusive ({} given)".format(
                gain
            ))
        else:
            g = gain

        # Directivity must be a multiple of 45
        if directivity is None:
            d = 0
        elif type(directivity) is int:
            if directivity % (360/8) != 0.0:
                raise ValueError("Directivity must be a multiple of 45 ({} given)".format(
                    directivity
                ))
            else:
                d = int(directivity / 45)
        else:
            raise TypeError(
                "Directivity must either be of type 'int' or None ({} given)".format(
                    type(directivity)
                ))

        return f"{s}{h}{g}{d}"

    @staticmethod
    def decode_nrq(nrq: str) -> Tuple[Optional[Union[float, str]], Optional[int], Optional[int]]:
        """
        Parse an NRQ (Number/Range/Quality) value and return the individual values.

        :param str nrq: an NRQ value

        For direction-finding reports, the NRQ value provides:-
         * The number of hits per period relative to the length of the time period (as a percentage)
         * The range (in miles)
         * The bearing accuracy (in degrees)

        NRQ values are 3 digits long, and all digits from 0 to 9. An 'N' value of 0 implies that the
        rest of the values are meaningless. An 'N' value of 9 indicates that the report is manual.

        The bearing accuracy represents the degree of accuracy, so a 'Q' value of 3 is 64, meaning
        that the accuracy is to less than 64 degrees.

        See APRS 1.01 C7 P30
        """

        # Ensure this is a valid NRQ value
        if not re.match(r'^[0-9]{3}$', nrq):
            raise ValueError(f"Invalid NRQ value: {nrq}")

        # Decode the number of hits
        # We can get this by dividing 100 / 8, then multiplying by the value
        if nrq[0] == "0":
            number = None
            rng = None
            quality = None
        else:
            if nrq[0] == "9":
                number = "manual"
                logger.debug(f"NRQ value is manually reported")
            else:
                number = (100/8) * int(nrq[0])
                logger.debug(f"Number of hits is {number}%")

            # The range is 2 to the power of the range value
            rng = 2 ** int(nrq[1])
            logger.debug(f"Range is {rng} miles")

            # The quality is in degrees. 9 through 3 double the value each time, starting at 1.
            # 2 and 1 are 120 and 240 respectively. 0 implies a useless quality level.
            if int(nrq[2]) > 2:
                quality = 2 ** (9-int(nrq[2]))
            elif int(nrq[2]) == 2:
                quality = 120
            elif int(nrq[2]) == 1:
                quality = 240
            else:
                quality = None

            if quality:
                logger.debug(f"Bearing accuracy is < {quality} degrees")
            else:
                logger.debug("Bearing accuracy is useless")

        return (number, rng, quality)

    @staticmethod
    def generate_passcode(callsign: str) -> str:
        """
        Generate an APRS-IS passcode for a given callsign.

        :param str callsign: a callsign, with or without an SSID

        This will generate an APRS-IS passcode for a callsign, after stripping the SSID (if given).
        """
        # Initialise the passcode with 0x73e2
        passcode = 0x73e2

        # Strip the SSID off the callsign, and convert it to uppercase
        base_call = callsign.split("-")[0].upper()

        # The algorithm XORs each character of the callsign with the passcode, alternating between
        # shifting the ASCII value 8 bits to the left
        high = True
        for c in base_call:
            if high:
                passcode ^= (ord(c) << 8)
                high = False
            else:
                passcode ^= ord(c)
                high = True

        # Return the passcode as a string
        return str(passcode)

    @classmethod
    def validate_passcode(cls, callsign: str, passcode: str) -> bool:
        """
        Validate an APRS-IS passcode for a given callsign.

        :param str callsign: a callsign, with or without an SSID
        :param str passcode: a passcode

        This will generate a passcode for the given callsign, and compare it to the provided
        passcode.
        """
        # Strip the SSID off the callsign, and convert it to uppercase
        base_call = callsign.split("-")[0].upper()

        # Generate a passcode for the given callsign
        generated_passcode = cls.generate_passcode(base_call)

        # Check to see if it matches the passcode we generated
        if passcode == generated_passcode:
            # Passcode matches
            return True
        else:
            # Passcode does not match
            return False



#!/usr/bin/env python

import logging
import re

from typing import List, Optional, Union
from enum import Enum

from .exceptions import ParseError

# Set up logging
logger = logging.getLogger(__name__)


class QConstruct(Enum):
    """
    Enum for specifying q constructs.
    """
    qAC = "qAC"
    qAX = "qAX"
    qAU = "qAU"
    qAo = "qAo"
    qAO = "qAO"
    qAS = "qAS"
    qAr = "qAr"
    qAR = "qAR"
    qAZ = "qAZ"
    qAI = "qAI"


class Station:
    """
    Class for describing a station, with an optional SSID
    """
    _callsign = None

    def __init__(self, callsign: str, ssid: Union[str, int] = None):
        """
        Create a new station.

        :param str callsign: a callsign, with or without an SSID
        :param str/int ssid: an SSID

        This will create a new :class:`Station` object, which describes a station - with or without
        an SSID.

        A station with an SSID can either be specified in a single argument (with ``callsign``) by
        appending the SSID to it, or by setting ``ssid``. Both ways are handled the same internally.

        **Note**: stations with SSIDs other than the numbers 1-15 are only valid on APRS-IS, and
        cannot be used over AX.25.
        """

        # 'ssid' must be set first, since it can be updated if we're passed a callsign with an SSID
        self.ssid = ssid
        self.callsign = callsign

    @property
    def callsign(self) -> str:
        """Get the callsign of the station"""
        return self._callsign

    @callsign.setter
    def callsign(self, value: str):
        """Set the callsign of the station (with or without an SSID)"""
        logger.debug(value)
        # Ensure we're being given a str
        if type(value) is not str:
            raise TypeError("Callsign must be of type 'str' ({} given)".format(type(value)))

        callsign_ssid_regex = '^[A-Za-z0-9]+-[A-Za-z0-9]{1,2}$'
        callsign_only_regex = '^[A-Za-z0-9]+$'

        if re.match(callsign_ssid_regex, value):
            # We have been given a callsign and an SSID
            (self._callsign, self.ssid) = value.split("-")

        elif re.match(callsign_only_regex, value):
            # We have only been given a callsign
            self._callsign = value

        else:
            raise ValueError("Callsign is invalid")

    @property
    def ssid(self) -> str:
        """Get the SSID of the station"""
        return self._ssid

    @ssid.setter
    def ssid(self, value: Union[str, int]):
        """Set the SSID of the station"""
        # Allow None
        if value is None:
            self._ssid = None

        # Ensure we're being given a str or an int
        elif type(value) is str:
            if re.match(r'^[A-Za-z0-9]{1,2}$', value):
                # Valid SSID
                if re.match(r'^[0-9]{1,2}', value):
                    # Can we convert it to an int?
                    try:
                        if int(value) == 0:
                            self._ssid = None
                        elif int(value) <= 15:
                            self._ssid = int(value)
                        else:
                            self._ssid = value

                    except ValueError:
                        # Not an integer
                        self._ssid = value
                else:
                    self._ssid = value
            else:
                raise ValueError("SSID is invalid")

        elif type(value) is int:
            if value == 0:
                self._ssid = None
            elif 1 <= value <= 15:
                self._ssid = value
            else:
                raise ValueError("SSID is invalid")

        else:
            raise TypeError("SSID must be of type 'str', 'int' or 'None' ({} given)".format(
                type(value)
            ))

    @property
    def is_valid_ax25(self) -> bool:
        """Get whether this station is a valid AX.25 address"""
        if len(self.callsign) > 6:
            # Callsign is longer than 6 characters
            return False

        if type(self.ssid) is int:
            return True
        elif self.ssid is None:
            return True
        else:
            return False

    def __str__(self) -> str:
        if self.ssid:
            return f"{self.callsign}-{self.ssid}"
        else:
            return self.callsign

    def __repr__(self) -> str:
        return "<Station: {}>".format(
            str(self)
        )


class PathHop:
    """
    Class for describing a single path hop.
    """
    def __init__(self, hop: Union[str, Station, QConstruct], used: bool = False):
        """
        Create a new path hop.

        :param str/Station/QConstruct hop: a path hop as either a ``str`` (with or without a
        trailing ``*`` to denote a used hop), a :class:`Station` or a :class:`Station`.
        :param bool used: whether the hop has been used up or not

        This will create a new :class:`PathHop` object, which describes a single hop in a path.
        A number of these can be grouped together in a single :class:`Path` object.

        A used hop can be either be specified in a single argument (with ``hop``) by appending
        a ``*`` to the end of the station, or by setting ``used`` to ``True``. Both ways are handled
        the same internally.
        """

        # 'used' must be set first, since it can be updated if we're only passed a station with a
        # '*'
        self.used = used
        self.hop = hop

    @property
    def hop(self) -> Union[Station, QConstruct]:
        """Get the hop"""
        return self._hop

    @hop.setter
    def hop(self, value: Union[str, Station, QConstruct]):
        """Set the hop"""
        # Check for a Station object
        if type(value) is Station:
            self._hop = value

        # Check for QConstruct object
        elif type(value) is QConstruct:
            self._hop = value

        # Parse the hop as a string
        elif type(value) is str:
            # Check for a q construct
            if value[0:2] == "qA":
                try:
                    self._hop = QConstruct(value=value)
                except (KeyError, ValueError):
                    raise ParseError("Invalid q construct: {}".format(value))

            # Check for a trailing *
            elif value[-1] == "*":
                self.used = True
                self._hop = Station(callsign=value[:-1])

            else:
                self.used = False
                self._hop = Station(callsign=value)
        else:
            raise TypeError(
                "Station must be of type 'str', 'Hop' or 'QConstruct' ({} given)".format(
                    type(value)
                )
            )

    @property
    def used(self) -> bool:
        """Get whether the hop has been used or not"""
        return self._used

    @used.setter
    def used(self, value: bool):
        """Set whether the hop has been used or not"""
        if type(value) is bool:
            self._used = value
        else:
            raise TypeError(
                "Used status must be of type 'bool' ({} given)".format(type(value))
            )

    def __str__(self) -> str:
        if self.used:
            return "{}*".format(str(self.hop))
        else:
            if type(self.hop) is QConstruct:
                return str(self.hop.value)
            else:
                return str(self.hop)

    def __repr__(self) -> str:
        return "<PathHop: {}>".format(str(self))


class Path:
    """
    Class for describing a path.
    """
    def __init__(self, path: Optional[Union[str, List[PathHop]]]):
        """
        Create a new path.

        :param str/List[PathHop] path: either a string representing a path, or a :term:`list` of
            :class:`PathHop` objects
        """
        self._path_hops = []
        if path:
            self.path = path

    @property
    def path(self) -> str:
        return ",".join([str(path_hop) for path_hop in self._path_hops])

    @path.setter
    def path(self, value: str):
        if type(value) is str:
            # Split the path
            self._path_hops = [PathHop(hop=path_hop) for path_hop in value.split(",")]
        else:
            raise TypeError("Path must be of type 'str' ({} given)".format(type(value)))

    @property
    def hops(self) -> list:
        return self._path_hops

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return "<Path: {}>".format(str(self))



#!/usr/bin/env python

import re
import logging

from ..exceptions import ParseError
from ..utils import APRSUtils
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class StationCapabilityPacket(GenericPacket):
    """
    Class to represent station capability packet.

    See APRS 1.01 C15 P77
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status_message = None
        self._maidenhead_locator = None
        self.capabilities = []

    def _parse(self) -> bool:
        """ TODO """
        capabilities = self._info.split(",")

        for capability in capabilities:
            if "=" in capability:
                (token, value) = capability.split("=")
                self.capabilities.append((token, value))
            else:
                self.capabilities.append((capability,))

        return True

    def __repr__(self):
        if self.source:
            return "<StationCapabilityPacket: {}>".format(self.source)
        else:
            return "<StationCapabilityPacket>"



#!/usr/bin/env python

import logging
import json
import enum
import re

from typing import Union
from geopy.point import Point
from datetime import datetime
from bitstring import Bits

from ..components import Path, Station
from ..exceptions import GenerateError

# Set up logging
logger = logging.getLogger(__name__)


class PacketJSONEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            if type(o) is datetime:
               return o.strftime('%c')
            elif type(o) is Bits:
                return o.bin
            else:
                j = {}
                for a, v in o.__dict__.items():
                    name = re.sub(r'^_', '', a)
                    if type(v) is Point:
                        j[name] = {
                            "latitude": v.latitude,
                            "longitude": v.longitude,
                            "altitude": v.altitude
                        }
                    elif type(v) is not enum.EnumMeta:
                        j[name] = v

                return j

        except:
            breakpoint()
            raise
            pass


class GenericPacket:
    """
    Generic base class for APRS packets.

    This is the base class for representing an APRS packet. Anything that is common to all packets
    is defined in here.
    """

    def __init__(self, source: str = None, destination: str = None, path: str = None,
                 data_type_id: str = None, info: str = None, symbol_table: str = None,
                 symbol_id: str = None):
        # list to hold the path hops
        self._path_hops = []

        self._raw = None

        # Set source, destination, path and info (if given)
        self.source = source
        self.destination = destination
        self.path = path
        self.data_type_id = data_type_id
        self.symbol_table = symbol_table
        self.symbol_id = symbol_id

        # TODO - info is a bit messed up
        self._info = info

        self.checksum = None
        self.timestamp = None
        self.timestamp_type = None

    @property
    def source(self) -> Station:
        """Get the source address of the packet"""
        return self._source

    @source.setter
    def source(self, value: Union[str, Station]):
        """Set the source address of the packet"""
        if type(value) is str:
            # Passed a str
            if len(value) <= 9:
                self._source = value
            else:
                raise ValueError("Source must be a maximum of 9 characters")
        elif type(value) is Station:
            # Passed a Station
            self._source = value
        elif value is None:
            # Passed None
            self._source = None
        else:
            raise TypeError("Source must either be 'str' or 'Station' ({} given)".format(
                type(value)
            ))

    @property
    def destination(self) -> Station:
        """Get the destination address of the packet"""
        return self._destination

    @destination.setter
    def destination(self, value: Union[str, Station]):
        """Set the destination address of the packet"""
        if type(value) is str:
            # Passed a str
            if len(value) <= 9:
                self._destination = value
            else:
                raise ValueError("Destination must be a maximum of 9 character")
        elif type(value) is Station:
            # Passed a Station
            self._destination = value
        elif value is None:
            # Passed None
            self._destination = None
        else:
            raise TypeError("Destination must be of type 'str' and a maximum of 9 characters")

    @property
    def path(self) -> Path:
        """Get the path of the packet"""
        return self._path

    @path.setter
    def path(self, value: Union[str, Path]):
        """Set the path for the packet"""
        if type(value) is str:
            self._path = Path(path=value)
        elif type(value) is Path:
            self._path = value
        elif value is None:
            self._path = None
        else:
            raise TypeError("Path must be of type 'str' or 'Path' ({} given)".format(type(value)))

    @property
    def info(self) -> str:
        """Get the info field of the packet"""
        return self._info

    @info.setter
    def info(self, value: str):
        """Set the info field of the packet"""
        self._info = value

    @property
    def timestamp(self) -> str:
        """Get the timestamp of the packet"""
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: str):
        """Set the timestamp of the packet"""
        self._timestamp = value

    @property
    def timestamp_type(self) -> str:
        """Get the timestamp type of the packet"""
        return self._timestamp_type

    @timestamp_type.setter
    def timestamp_type(self, value: str):
        """Set the timestamp type of the packet"""
        self._timestamp_type = value

    @property
    def data_type_id(self) -> str:
        """Get the data type ID of the packet"""
        return self._data_type_id

    @data_type_id.setter
    def data_type_id(self, value: str):
        """Set the data type ID of the packet"""
        self._data_type_id = value

    @property
    def symbol_table(self) -> str:
        """Get the symbol table of the packet"""
        return self._symbol_table

    @symbol_table.setter
    def symbol_table(self, value: str):
        """Set the symbol table of the packet"""
        self._symbol_table = value

    @property
    def symbol_id(self) -> str:
        """Get the symbol ID of the packet"""
        return self._symbol

    @symbol_id.setter
    def symbol_id(self, value: str):
        """Set the symbol ID of the packet"""
        self._symbol = value

    @property
    def raw(self) -> str:
        if self._raw:
            return "{}>{}:{}:{}".format(
                self._raw.source, self._raw.destination, self._raw.path, self._raw.information
            )

    def _parse(self) -> bool:
        return True

    def generate(self):
        """
        Generate an APRS packet based on the current object's properties.
        """
        output = ""
        if self.source is not None:
            output += f"{self.source}>"
        else:
            raise GenerateError("Missing source address")

        if self.destination is not None:
            output += f"{self.destination},"
        else:
            raise GenerateError("Missing destination address")

        if self.path is not None:
            output += f"{self.path}:"
        else:
            raise GenerateError("Missing path")

        info = self.info

        if self.data_type_id is not None:
            output += f"{self.data_type_id}"
        else:
            raise GenerateError("Missing data type ID")

        if info is not None:
            output += f"{info}"
        else:
            raise GenerateError("Missing info")

        return output

    def to_json(self):
        j = PacketJSONEncoder().encode(self)

        return j

    def __repr__(self):
        if self.source:
            return "<GenericPacket: {}>".format(self.source)
        else:
            return "<GenericPacket>"



#!/usr/bin/env python

import logging

from typing import Tuple

from ..exceptions import ParseError
from ..utils import APRSUtils
from .position import PositionPacket

# Set up logging
logger = logging.getLogger(__name__)


class MICEPacket(PositionPacket):
    """
    Class to represent Mic-E encoded position packets.

    Mic-E packets encode position data in both the APRS destination and information fields, enabling
    a large amount of information to be conveyed in a single packet.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def _decode_latitude(destination: str) -> Tuple[float, int, bool, str, int, int, int, bool]:
        """
        Decode the latitude.

        :param str destination: the destination field of a packet

        Mic-E packets encode the latitude in the destination field, along with the north/south,
        east/west, longitude offset and message identifier. In addition, the SSID can be used to
        denote a generic APRS digipeater path.

        For more information, seeee APRS 1.01 C10 P43.
        """

        latitude = ""
        north_south = ""
        east_west = ""
        # TODO - handle status types
        message_a = ""
        message_b = ""
        message_c = ""
        message_custom = False
        lng_offset = False

        # Iterate over each character of the destination address
        # TODO - A-K not used for 4-6
        count = 0
        for i in destination[0:6]:
            count += 1
            if 48 <= ord(i) <= 57:
                # 0-9 are used as-is
                latitude += i
                enc_set = 1
            elif 65 <= ord(i) <= 74:
                latitude += str((ord(i) - 65))
                enc_set = 2
            elif 80 <= ord(i) <= 89:
                latitude += str((ord(i) - 80))
                enc_set = 3
            elif i in "KLZ":
                latitude += " "
                if i == "K":
                    enc_set = 2
                elif i == "L":
                    enc_set = 1
                elif i == "Z":
                    enc_set = 3
            else:
                raise ParseError("Unexpected character in Mic-E destination field")

            if count == 1:
                # The first field is changed depending on the message a-bit
                if enc_set == 1:
                    message_a = 0
                elif enc_set == 2:
                    message_a = 1
                    message_custom = True
                elif enc_set == 3:
                    message_a = 1
            elif count == 2:
                # The second field is changed depending on the message b-bit
                if enc_set == 1:
                    message_b = 0
                elif enc_set == 2:
                    message_b = 1
                    message_custom = True
                elif enc_set == 3:
                    message_b = 1
            elif count == 3:
                # The third field is changed depending on the message c-bit
                if enc_set == 1:
                    message_c = 0
                elif enc_set == 2:
                    message_c = 1
                    message_custom = True
                elif enc_set == 3:
                    message_c = 1
            elif count == 4:
                # The fourth field is changed to denote north/south
                if enc_set == 1:
                    north_south = "S"
                else:
                    north_south = "N"

                # Append a . to the latitude here, too
                latitude += "."
            elif count == 5:
                # The fifth field is changed to denote the longitude offset
                if enc_set == 1:
                    lng_offset = False
                else:
                    lng_offset = True
            elif count == 6:
                # The sixth field is changed to denote east/west
                if enc_set == 1:
                    east_west = "E"
                else:
                    east_west = "W"

        # TODO
        if len(destination) > 6:
            logger.debug("Mic-E destination has SSID: {}".format(destination.split('-')[1]))

        # Now that we have an uncompressed latitude, we can decode it like a standard uncompressed
        # packet
        try:
            decoded_latitude, ambiguity = APRSUtils.decode_uncompressed_latitude(
                "{}{}".format(latitude, north_south)
            )

        except ValueError as e:
            raise ParseError(e)

        logger.debug(
            "After destination field decoding, latitude is {}{} ({} - {}), longitude offset is {}, "
            "direction is {}, msg a/b/c is {}/{}/{}, msg custom is {}".format(
                latitude, north_south, decoded_latitude, ambiguity, lng_offset, east_west,
                message_a, message_b, message_c, message_custom
            ))

        # Return the decoded latitude, ambiguity, longitude offset, east/west direction and message
        # bits
        return (decoded_latitude, ambiguity, lng_offset, east_west, message_a, message_b,
                message_c, message_custom)

    @staticmethod
    def _decode_longitude(info: str, lng_offset: bool, east_west: str) -> float:
        """
        Decode the longitude.

        :param str info: the information field of a packet, minus the initial data type identifier
        :param bool lng_offset: whether the longitude offset should be applied or not
        :param str east_west: a single character (``E`` or ``W``) denoting an east or west latitude

        The longitude is stored in the 3 characters of the info field immediately following the data
        type identifier. Combined with the east/west and longitude offset values from decoding the
        latitude, we can decode the longitude.

        See also APRS 1.01 C10 P47.
        """

        logger.debug("Info input: {}".format(info))
        # The degrees, minutes and hundreths of minutes are obtained by subtracting 28 from the
        # ASCII values of the 3 characters

        try:
            lng_deg = ord(info[0]) - 28
            lng_min = ord(info[1]) - 28
            lng_hmin = ord(info[2]) - 28

        except IndexError as e:
            raise ParseError("Invalid longitude: {}".format(e))

        # If the longitude offset is set, apply it to the degrees value
        if lng_offset:
            lng_deg += 100

        # Next, if the value is between 180 and 189, subtract 80. If it's between 190 and 199,
        # subtract 190.
        if 180 <= lng_deg <= 189:
            lng_deg -= 80
        elif 190 <= lng_deg <= 199:
            lng_deg -= 190

        # If the minutes value is more than 60, subtract 60
        if lng_min >= 60:
            logger.debug("> 60 subtracted from lng_min")
            lng_min -= 60

        # Now that we have an uncompressed longitude, we can decode it
        try:
            longitude = APRSUtils.decode_uncompressed_longitude("{}{}.{}{}".format(
                str(lng_deg).zfill(3),
                str(lng_min).zfill(2),
                str(lng_hmin).zfill(2),
                east_west
            ))

        except ValueError as e:
            raise ParseError(e)

        logger.debug(
            "Longitude is {} {} {} ({})".format(lng_deg, lng_min, lng_hmin, longitude)
        )

        return longitude

    @staticmethod
    def _decode_speed_and_course(info: str):
        """
        Decode the speed and course.

        :param str info: the information field of a packet, minus the initial data type identifier

        The 3 characters after the longitude characters define the speed and course of the station.

        See also APRS 1.01 C10 P49.
        """

        try:
            # For each of the 3 characters, subtract 28 from the ASCII value
            sp = ord(info[3]) - 28
            dc = ord(info[4]) - 28
            se = ord(info[5]) - 28
        except IndexError:
            raise ParseError("Couldn't parse speed/course in Mic-E packet")

        # The speed is in knots, and is obtained by multiplying sp by 10, and adding the quotient of
        # dc divided by 10
        # TODO - handle other speed encoding scheme (see C10 P50)
        speed = (sp * 10) + int(dc / 10)

        # The course is in degrees, and is obtained by multiplying the remainder of dc divided by 10
        # by 100, and then adding the value of se
        course = ((dc % 10) * 100) + se

        # If the speed value is greater than or equal to 800, subtract 800 from it
        if speed >= 800:
            speed -= 800

        # If the course value is greater than or equal to 400, subtract 400 from it
        if course >= 400:
            course -= 400

        logger.debug("Speed is {} knots, course is {} degrees".format(speed, course))

        return (speed, course)

    def _parse(self) -> bool:
        """
        Parse a Mic-E packet.

        The parsed and decoded values are stored in the current object.
        """

        # Decode the latitude, ambiguity, the longitude offset, the east/west direction and the
        # message bits
        (self.latitude, self.ambiguity, lng_offset, east_west, message_a, message_b, message_c,
         message_custom) = self._decode_latitude(self.destination)

        # Decode the longitude, using the longitude offset and east/west direction from the previous
        # step
        self.longitude = self._decode_longitude(self._info, lng_offset, east_west)

        # Decode the speed and course from the info field
        (self.speed, self.course) = self._decode_speed_and_course(self._info)

        # Parse the symbol table and symbol from the info field
        try:
            self.symbol_id = self._info[6]
            logger.debug("Symbol ID is {}".format(self.symbol_id))
        except IndexError:
            raise ParseError("Missing symbol ID", self)

        try:
            self.symbol_table = self._info[7]
            logger.debug("Symbol table is {}".format(self.symbol_table))
        except IndexError:
            raise ParseError("Missing symbol table", self)

        # Next comes either the status text or telemetry (C10 P54)
        # Telemetry is indicated by setting the first character of the status field to dec 44
        # (a ',') or the unprintable 29.
        if len(self._info) >= 10:
            if ord(self._info[8]) == 44 or ord(self._info[8]) == 29:
                logger.debug("Packet contains telemetry data")
                # TODO
            else:
                logger.debug("Packet contains status text")
                status_text = self._info[9:]

                # The status text can contain an altitude value (C10 P55)
                # It's indicated by the first 4 characters, with the 4th being a '}'
                if len(status_text) >= 4 and status_text[3] == "}":
                    logger.debug("Status text contains altitude data")

                    # Decode the altitude
                    altitude = (
                        (ord(status_text[0])-33) * 91**2 +
                        (ord(status_text[1])-33) * 91 +
                        (ord(status_text[2])-33)
                    ) - 10000

                    self.altitude = altitude
                    logger.debug("Altitude is {}m".format(altitude))

                    # The remainder is the comment
                    self.comment = status_text[4:]
                    logger.debug("Comment is {}".format(self.comment))

                else:
                    self.comment = status_text
        else:
            logger.debug("Packet contains no further information.")

        return True

    def __repr__(self):
        if self.source:
            return "<MICEPacket: {}>".format(self.source)
        else:
            return "<MICEPacket>"



#!/usr/bin/env python

import re
import logging

from hashlib import md5
from collections import namedtuple
from datetime import datetime

from .exceptions import ParseError, UnsupportedError
from .packets.generic import GenericPacket
from .packets.beacon import BeaconPacket
from .packets.position import PositionPacket
from .packets.status import StatusPacket
from .packets.mice import MICEPacket
from .packets.object import ObjectPacket
from .packets.message import MessagePacket
from .packets.telemetry import TelemetryPacket
from .packets.telemetry_definition import TelemetryParameterNamePacket, TelemetryUnitLabelPacket,\
    TelemetryEquationCoefficientsPacket, TelemetryBitSenseProjectNamePacket
from .packets.station_capability import StationCapabilityPacket

__version__ = "0.3.3"

# Set up logging
logger = logging.getLogger(__name__)

# Named tuple to hold original raw values
Raw = namedtuple('Raw', ['source', 'destination', 'path', 'information'])

BEACON_ADDRESSES = [
    'AIR', 'ALL', 'AP', 'BEACON', 'CQ', 'GPS', 'DF', 'DGPS', 'DRILL', 'DX', 'ID', 'JAVA', 'MAIL',
    'MICE', 'QST', 'QTH', 'RTCM', 'SKY', 'SPACE', 'SPC', 'SYM', 'TEL', 'TEST', 'TLM', 'WX', 'ZIP'
]


def _handle_err(exception: Exception, packet: str, strict_mode: bool = True):
    if not strict_mode:
        logger.warning("Returning generic packet for: {}".format(packet))
    else:
        raise exception


class APRS:
    """
    Main APRS class.

    This class provides functions for parsing and decoding different kinds of APRS packets.
    Packet type-specific functions are defined under the different :class:`APRSPacket` subclasses.
    """

    @staticmethod
    def parse(packet: str = None, timestamp: datetime = None,
              strict_mode: bool = True) -> GenericPacket:
        """
        Parse an APRS packet, and return a subclass of :class:`APRSPacket` appropriate for the
        packet type.

        :param str packet: a raw packet
        :param datetime timestamp: an (optional) timestamp indicating when the packet arrived

        Given a raw packet, this function will return a object that is a subclass of
        :class:`APRSPacket`.
        """

        try:
            # Parse out the source, destination, path and information fields
            (source, destination, path, data_type_id, info) = re.match(
                r'([\w\d\-]+)>([\w\d\-]+),([\w\d\-\*\,]+):(.)(.*)',
                packet
            ).groups()
        except AttributeError:
            _handle_err(
                ParseError("Could not parse packet details", packet),
                packet,
                strict_mode
            )
            p = GenericPacket()

        # Create a checksum, to provide a quick comparison against other packets
        checksum = md5((source + info).encode()).hexdigest()
        logger.debug("Packet checksum is {}".format(checksum))

        logger.debug("Raw packet: {}".format(packet))

        # Do some basic sanity checking
        # The source and destination fields should be a maximum of 9 characters
        logger.debug("Destination length is {}".format(len(destination)))
        if len(source) > 9:
            _handle_err(
                ParseError("Source address is longer than 9 characters", packet),
                packet,
                strict_mode
            )
            p = GenericPacket()

        elif len(destination) > 9:
            _handle_err(
                ParseError("Destination address is longer than 9 characters", packet),
                packet,
                strict_mode
            )
            p = GenericPacket()

        # The destination field should be upper case
        if not re.match(r'^[A-Z0-9]{1,6}(\-[0-9]{1,2})?$', destination):
            _handle_err(
                ParseError("Destination address is invalid", packet),
                packet,
                strict_mode
            )
            p = GenericPacket()

        # Check for common beacon destinations
        if destination in BEACON_ADDRESSES:
            logger.debug("Packet is a beacon packet")
            p = BeaconPacket()

        elif data_type_id in '!/=@':
            logger.debug("Packet is a position packet")
            p = PositionPacket()

            if len(info) < 4:
                # TODO - handle this properly
                _handle_err(
                    ParseError("Packet is too short."),
                    packet,
                    strict_mode
                )
                p = GenericPacket()

        elif data_type_id in "`'":
            logger.debug("Packet is a Mic-E packet")
            p = MICEPacket()

        elif data_type_id == ";":
            logger.debug("Packet is an object packet")
            p = ObjectPacket()

        elif data_type_id == ")":
            _handle_err(
                UnsupportedError("Unsupported data type: '<' (Item Report) - C11 P57"),
                packet,
                strict_mode
            )
            p = GenericPacket()

        # Check for the presence of PARM, UNIT, EQNS or BITS, indicating a message defining
        # telemetry data. The APRS spec places a 67-character limit on the message field, but
        # it's common to see telemetry definitions exceed this
        elif re.search(r'::[A-Za-z0-9\-]+\s?:(PARM|UNIT|EQNS|BITS)\.', packet):
            definition_type = re.search(
                r'::[A-Za-z0-9\-]+\s?:(PARM|UNIT|EQNS|BITS)\.',
                packet
            ).groups()[0]

            if definition_type == "PARM":
                p = TelemetryParameterNamePacket()

            elif definition_type == "UNIT":
                p = TelemetryUnitLabelPacket()

            elif definition_type == "EQNS":
                p = TelemetryEquationCoefficientsPacket()

            elif definition_type == "BITS":
                p = TelemetryBitSenseProjectNamePacket()

            else:
                _handle_err(
                    ParseError("Invalid telemetry definition type: {}".format(definition_type)),
                    packet,
                    strict_mode
                )
                p = GenericPacket()

        elif data_type_id == ":":
            logger.debug("Packet is a message packet")
            p = MessagePacket()

        elif data_type_id == "T":
            logger.debug("Packet is a telemetry packet")
            p = TelemetryPacket()

        elif data_type_id == ">":
            logger.debug("Packet is a status packet")
            p = StatusPacket()

        elif data_type_id == "<":
            logger.debug("Packet is a station capability packet")
            p = StationCapabilityPacket()

        elif data_type_id == "$":
            _handle_err(
                UnsupportedError("Unsupported data type: '$' (Raw NMEA Position Report)"),
                packet,
                strict_mode
            )
            p = GenericPacket()

        elif data_type_id == "_":
            _handle_err(
                UnsupportedError("Unsupported data type: '_' (Positionless Weather Report"),
                packet,
                strict_mode
            )
            p = GenericPacket()

        elif data_type_id == "*":
            _handle_err(
                UnsupportedError("Unsupported data type: '*' (Complete Weather Report"),
                packet,
                strict_mode
            )
            p = GenericPacket()

        elif 0 <= info.find('!') <= 40:
            # As per APRS 1.01 C5 P18, position-without-timestamp packets may have the '!' located
            # anywhere up to the 40th character in the information field. If we're here, test for
            # that now.
            logger.debug("Found ! in information field, parsing as position packet")
            p = PositionPacket()

            # Store the offset
            p._offset = info.find('!')

            # Because we normally assume the first character of the info field is the data type ID,
            # update the info field to include it
            info = data_type_id + info

            # Set the data type ID
            data_type_id = '!'

        else:
            _handle_err(
                UnsupportedError("Unknown data type: {} (raw: {})".format(data_type_id, packet)),
                packet,
                strict_mode
            )
            # Return a generic APRS Packet with the basic info
            p = GenericPacket()

        # Set the source, destination, path and information fields, along with the checksum, data
        # type ID and the raw packet
        p.source = source
        p.destination = destination
        p.path = path
        p._info = info
        p.checksum = checksum
        p.data_type_id = data_type_id
        p._ts = timestamp

        # Add the raw values to the packet object
        p._raw = Raw(source=source, destination=destination, path=path, information=info)

        # Call the packet-specific parser
        try:
            p._parse()

        except ParseError as e:
            _handle_err(
                e, packet, strict_mode
            )

            p = GenericPacket()

            p.source = source
            p.destination = destination
            p.path = path
            p._info = info
            p.checksum = checksum
            p.data_type_id = data_type_id
            p._ts = timestamp

            # Add the raw values to the packet object
            p._raw = Raw(source=source, destination=destination, path=path, information=info)

        # Return the packet object
        return p



#!/usr/bin/env python

import re
import logging

from ..exceptions import ParseError
from ..utils import APRSUtils
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class StatusPacket(GenericPacket):
    """
    Class to represent status report packets.

    As per APRS 1.01, status reports "announce(s) the station's current mission or any other single
    line status to everyone". The data type identifier for status reports is ">".

    See APRS 1.01 C16 P80
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status_message = None
        self._maidenhead_locator = None

    @property
    def status_message(self) -> str:
        """Get the status message"""
        return self._status_message

    @status_message.setter
    def status_message(self, value: str):
        """Set the status message"""
        self._status_message = value

    @property
    def maidenhead_locator(self) -> str:
        """Get the Maidenhead locator (if set)"""
        return self._maidenhead_locator

    @maidenhead_locator.setter
    def maidenhead_locator(self, value: str):
        """Set the Maidenhead locator"""
        self._maidenhead_locator = value

    def _parse(self) -> bool:
        """
        Parse a status report packet.

        Status reports contain a status message and optionally a timestamp or Maidenhead locator
        (but not both), and/or a beam heading and ERP value.

        Parse and decoded values are stored in the current object.
        """

        # Maidenhead locators, if present, must be first, and can be 4 or 6 characters.
        # They must also include a symbol table and symbol code (P81).
        # Therefore, if the status report matches [A-Z]{2}[0-9]{2}([A-Z]{2})? and a valid symbol
        # table and id, then it's a Maidenhead locator.
        # Valid symbol table IDs: /, \, 0-9, A-Z (C20 P91)
        mh_4 = None
        mh_6 = None

        if len(self._info) >= 6:
            mh_4 = self._info[0:6]
            logger.debug("Considering as Maidenhead locator: {}".format(mh_4))

        if len(self._info) >= 8:
            mh_6 = self._info[0:8]
            logger.debug("Considering as Maidenhead locator: {}".format(mh_6))

        if mh_6 is not None and re.match("[A-Z]{2}[0-9]{2}[A-Z]{2}[/\\\0-9A-Z].", mh_6):
            # Maidenhead locator (GGnngg)
            self.maidenhead_locator = mh_6[0:6]

            self.symbol_table = mh_6[6]
            self.symbol_id = mh_6[7]

            logger.debug("Status with Maidenhead locator {}, symbol {} {}".format(
                self.maidenhead_locator, self.symbol_table, self.symbol_id
            ))

            if len(self._info) != 8:
                # First character of the text must be " " (C16 P82)
                if self._info[8] != " ":
                    # TODO
                    raise ParseError("Status message is invalid", self)
                else:
                    self.status_message = self._info[9:]
                    logger.debug("Status message is {}".format(self.status_message))
            else:
                logger.debug("No status message")

        elif mh_4 is not None and re.match(r'[A-Z]{2}[0-9]{2}[/\\\0-9A-Z].', mh_4):
            # Maidenhead locator (GGnn)
            self.maidenhead_locator = mh_4[0:4]

            self.symbol_table = mh_4[4]
            self.symbol_id = mh_4[5]

            logger.debug("Status with Maidenhead locator {}, symbol {} {}".format(
                self.maidenhead_locator, self.symbol_table, self.symbol_id
            ))

            if len(self._info) != 6:

                # First character of the text must be " " (C16 P82)
                if self._info[6] != " ":
                    # TODO
                    raise ParseError("Status message is invalid", self)
                else:
                    self.status_message = self._info[7:]
                    logger.debug("Status message is {}".format(self.status_message))
            else:
                logger.debug("No status message")

        else:
            # Check for a timestamp
            if re.match("^[0-9]{6}z", self._info):
                try:
                    self.timestamp = APRSUtils.decode_timestamp(self._info[0:7])
                    # TODO Sanity check the timestamp type - status reports can only use zulu
                    # or local, so if hms is used, throw an error.
                    # if timestamp_type == 'h' and data_type_id == '>':
                    #     logger.error("Timestamp type 'h' cannot be used for status reports")
                    #     raise ParseError("Timestamp type 'h' cannot be used for status reports")
                    logger.debug("Status message timestamp is {}".format(self.timestamp))

                except ParseError:
                    self.timestamp = None

                self.status_message = self._info[7:]
                logger.debug("Status message is {}".format(self.status_message))

            else:
                self.status_message = self._info
                logger.debug("No timestamp found")
                logger.debug("Status message is {}".format(self.status_message))

        # Check for a beam heading and ERP
        # TODO - Parse and decode the beam and ERP values
        if self.status_message and re.match(r'[\^].{2}', self.status_message[-3:]):
            logger.debug("Status message heading and power: {}".format(self.status_message[-2:]))

        return True

    def __repr__(self):
        if self.source:
            return "<StatusPacket: {}>".format(self.source)
        else:
            return "<StatusPacket>"



#!/usr/bin/env python

import logging

# Set up logging
logger = logging.getLogger(__name__)


class ParseError(Exception):
    """
    Parsing exception
    """
    def __init__(self, message: str, packet=None):
        super().__init__(message)

        self.packet = packet


class GenerateError(Exception):
    """
    Generating exception
    """
    def __init__(self, message: str, packet=None):
        super().__init__(message)

        self.packet = packet


class UnsupportedError(Exception):
    """
    Thrown when packets are of an unsupported format
    """
    def __init__(self, message: str, packet=None):
        super().__init__(message)

        self.packet = packet



#!/usr/bin/env python

import re
import logging
import bitstring
from bitstring import Bits

from ..exceptions import ParseError
from ..utils import APRSUtils
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class TelemetryAnalogValue:
    _value = None

    def __init__(self, value):
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if type(value) is str:
            if value == "":
                value = None
            elif "." in value:
                value = round(float(value), 2)
            else:
                value = int(value)

        self._value = value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> int:
        return "<TelemetryAnalogValue: {}>".format(self.value)


class TelemetryDigitalValue:
    _value = None

    def __init__(self, value):
        self.value = value

    @property
    def value(self) -> Bits:
        return self._value

    @value.setter
    def value(self, value):
        if type(value) is str:
            if len(value) <= 8:
                self._value = Bits("0b{}".format(value))
            else:
                raise ValueError(
                    "Value must represent 8 bits or less ({} given)".format(value)
                )

        elif type(value) is int:
            self._value = Bits(uint=value, length=8)

    def __str__(self) -> str:
        return str(self.value.bin)

    def __repr__(self) -> int:
        return "<TelemetryDigitalValue: {}>".format(self.value)


class TelemetryPacket(GenericPacket):
    """
    Class to represent telemetry packets.

    See APRS 1.01 C13 P68
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._sequence_number = None
        self._av1 = None
        self._av2 = None
        self._av3 = None
        self._av4 = None
        self._av5 = None
        self._av5 = None
        self._dv = None
        self._comment = None

    @property
    def sequence_number(self) -> str:
        return self._sequence_number

    @sequence_number.setter
    def sequence_number(self, value: str):
        self._sequence_number = value

    @property
    def av1(self) -> TelemetryAnalogValue:
        return self._av1

    @av1.setter
    def av1(self, value):
        if type(value) is TelemetryAnalogValue:
            self._av1 = value
        else:
            self._av1 = TelemetryAnalogValue(value)

    @property
    def av2(self) -> TelemetryAnalogValue:
        return self._av2

    @av2.setter
    def av2(self, value):
        if type(value) is TelemetryAnalogValue:
            self._av2 = value
        else:
            self._av2 = TelemetryAnalogValue(value)

    @property
    def av3(self) -> TelemetryAnalogValue:
        return self._av3

    @av3.setter
    def av3(self, value):
        if type(value) is TelemetryAnalogValue:
            self._av3 = value
        else:
            self._av3 = TelemetryAnalogValue(value)

    @property
    def av4(self) -> TelemetryAnalogValue:
        return self._av4

    @av4.setter
    def av4(self, value):
        if type(value) is TelemetryAnalogValue:
            self._av4 = value
        else:
            self._av4 = TelemetryAnalogValue(value)

    @property
    def av5(self) -> TelemetryAnalogValue:
        return self._av5

    @av5.setter
    def av5(self, value):
        if type(value) is TelemetryAnalogValue:
            self._av5 = value
        else:
            self._av5 = TelemetryAnalogValue(value)

    @property
    def dv(self) -> TelemetryDigitalValue:
        return self._dv

    @dv.setter
    def dv(self, value):
        if type(value) is TelemetryDigitalValue:
            self._dv = value
        else:
            self._dv = TelemetryDigitalValue(value)

    @property
    def comment(self) -> str:
        return self._comment

    @comment.setter
    def comment(self, value: str):
        self._comment = value

    def _parse(self) -> bool:
        """
        Parse a telemetry packet.

        Parse and decoded values are stored in the current object.
        """

        try:
            (self.sequence_number, values) = re.match(r'#(MIC|\d+),?(.*)', self._info).groups()

        except AttributeError:
            # We didn't get any matches
            raise ParseError("Failed to parse telemetry packet.")

        try:
            (self.av1, self.av2, self.av3, self.av4, self.av5, remainder) = values.split(",", 5)

        except ValueError as e:
            # Failed to parse values
            raise ParseError("Failed to parse analog values: {}".format(e))

        # A comment, if it exists, is sometimes separated from the values by a space, and sometimes
        # by a comma. Complicating things further, sometimes the digital values aren't 8 characters
        # long.
        # Telemetry packets without the digital section BUT with a comment have also been seen. If
        # we catch an exception doing this, assume the whole remainder is a comment
        try:
            (self.dv, self.comment) = re.match(r'(\d+)[\s,]?(.*)', remainder).groups()

        except AttributeError:
            # No match, which means there's no digital values
            self.comment = remainder

        except ValueError:
            # This is likely due to an invalid digital value
            logger.warning("Invalid digital value in '{}', ignoring.".format(remainder))

        except bitstring.CreationError:
            # Failed to add the digital value, so skip it
            logger.warning("Invalid digital value in '{}', ignoring.".format(remainder))

        return True

    def __repr__(self):
        if self.source:
            return "<TelemetryPacket: {}>".format(self.source)
        else:
            return "<TelemetryPacket>"



#!/usr/bin/env python

import logging

from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class ObjectPacket(GenericPacket):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        if self.source:
            return "<ObjectPacket: {}>".format(self.source)
        else:
            return "<ObjectPacket>"

    def _parse(self) -> bool:
        return True



#!/usr/bin/env python

import re
import logging

from ..exceptions import ParseError, GenerateError
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class TelemetryDefinitionPacket(GenericPacket):
    """
    Class to represent APRS message packets.

    APRS message packets can be from one station to another, or general bulletins or announcements
    which are intended for a wider audience.

    See also APRS 1.01 C14 P71
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new :class:`MessagePacket` object.

        :param str addressee: the message addressee
        :param str message: the message
        """
        super().__init__(*args, **kwargs)

        # Set the data type ID
        self.data_type_id = ":"

    def _parse(self) -> bool:
        pass

    def __repr__(self):
        if self.source:
            return "<{}: {}>".format(self.__class__.__name__, self.source)
        else:
            return "<{}>".format(self.__class__.__name__)


class TelemetryParameterNamePacket(TelemetryDefinitionPacket):
    def _parse(self) -> bool:
        # If this is a message, then ':" MUST be in the 9th position (C14 P71)
        try:
            if self._info[9] != ":":
                raise ParseError(
                    "Invalid telemetry definition packet (missing : in 9th position)",
                    self
                )

        except IndexError:
            raise ParseError("Invalid telemetry definition packet (packet is too short)")

        # Get the values, ignoring the addressee portion. As per the spec (C13 P68) "The
        # message addressee is the callsign of the of the station transmitting the telemetry data".
        # We can also ignore the next 5 characters as they contain the definition type
        values = self._info[15:].split(",")

        # Fields
        fields = ['a1', 'a2', 'a3', 'a4', 'a5', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8']

        field_number = 0

        for value in values:
            setattr(self, fields[field_number], value)
            field_number += 1

        return True


class TelemetryUnitLabelPacket(TelemetryDefinitionPacket):
    def _parse(self) -> bool:
        # If this is a message, then ':" MUST be in the 9th position (C14 P71)
        try:
            if self._info[9] != ":":
                raise ParseError(
                    "Invalid telemetry definition packet (missing : in 9th position)",
                    self
                )

        except IndexError:
            raise ParseError("Invalid telemetry definition packet (packet is too short)")

        # Get the values, ignoring the addressee portion. As per the spec (C13 P68) "The
        # message addressee is the callsign of the of the station transmitting the telemetry data".
        # We can also ignore the next 5 characters as they contain the definition type
        values = self._info[15:].split(",")

        # Fields
        fields = ['a1', 'a2', 'a3', 'a4', 'a5', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8']

        try:
            field_number = 0
            for value in values:
                setattr(self, fields[field_number], value)
                field_number += 1

        except IndexError:
            raise ParseError("Invalid number of fields for telemetry definition packet.")

        return True


class TelemetryEquationCoefficientsPacket(TelemetryDefinitionPacket):
    def _parse(self) -> bool:
        # If this is a message, then ':" MUST be in the 9th position (C14 P71)
        try:
            if self._info[9] != ":":
                raise ParseError(
                    "Invalid telemetry definition packet (missing : in 9th position)",
                    self
                )

        except IndexError:
            raise ParseError("Invalid telemetry definition packet (packet is too short)")

        # Get the values, ignoring the addressee portion. As per the spec (C13 P68) "The
        # message addressee is the callsign of the of the station transmitting the telemetry data".
        # We can also ignore the next 5 characters as they contain the definition type
        values = self._info[15:].split(",")

        # Fields
        fields = [
            'a1_1', 'a1_2', 'a1_3',
            'a2_1', 'a2_2', 'a2_3',
            'a3_1', 'a3_2', 'a3_3',
            'a4_1', 'a4_2', 'a4_3',
            'a5_1', 'a5_2', 'a5_3'
        ]

        field_number = 0
        try:
            for value in values:
                setattr(self, fields[field_number], value)
                field_number += 1
        except IndexError:
            logger.warning("Equation coefficient specifies too many fields ({} > {})".format(
                len(values), len(fields)
            ))

        return True


class TelemetryBitSenseProjectNamePacket(TelemetryDefinitionPacket):
    def _parse(self) -> bool:
        # If this is a message, then ':" MUST be in the 9th position (C14 P71)
        try:
            if self._info[9] != ":":
                raise ParseError(
                    "Invalid telemetry definition packet (missing : in 9th position)",
                    self
                )

        except IndexError:
            raise ParseError("Invalid telemetry definition packet (packet is too short)")

        # Get the values, ignoring the addressee portion. As per the spec (C13 P68) "The
        # message addressee is the callsign of the of the station transmitting the telemetry data".
        # We can also ignore the next 5 characters as they contain the definition type
        values = self._info[15:].split(",")

        # Fields
        fields = ['b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'b8', 'project_title']

        field_number = 0
        for value in values:
            setattr(self, fields[field_number], value)
            field_number += 1

        return True



#!/usr/bin/env python

import logging

from ..exceptions import ParseError
from ..utils import APRSUtils
from .generic import GenericPacket

# Set up logging
logger = logging.getLogger(__name__)


class BeaconPacket(GenericPacket):
    """
    Class to represent beacon packets.

    See APRS 1.01 C4 P12
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._comment = None

    @property
    def comment(self) -> str:
        """Get the comment"""
        return self._comment

    @comment.setter
    def comment(self, value: str):
        """Set the comment"""
        self._comment = value

    def _parse(self) -> bool:
        """
        Parse a beacon packet.
        """

        self._comment = self._info
        logger.debug("Beacon packet comment is {}".format(self._comment))

        return True

    def __repr__(self):
        if self.source:
            return "<BeaconPacket: {}>".format(self.source)
        else:
            return "<BeaconPacket>"
